"""Class definitions for base aligner"""
from __future__ import annotations

import multiprocessing as mp
import os
import shutil
import time
from queue import Empty
from typing import List, Optional

import tqdm

from montreal_forced_aligner.abc import FileExporterMixin
from montreal_forced_aligner.alignment.mixins import AlignMixin
from montreal_forced_aligner.alignment.multiprocessing import (
    ExportTextGridArguments,
    ExportTextGridProcessWorker,
    PhoneAlignmentArguments,
    PhoneAlignmentFunction,
    WordAlignmentArguments,
    WordAlignmentFunction,
)
from montreal_forced_aligner.corpus.acoustic_corpus import AcousticCorpusPronunciationMixin
from montreal_forced_aligner.textgrid import export_textgrid, output_textgrid_writing_errors
from montreal_forced_aligner.utils import KaldiProcessWorker, Stopped

__all__ = ["CorpusAligner"]


class CorpusAligner(AcousticCorpusPronunciationMixin, AlignMixin, FileExporterMixin):
    """
    Mixin class that aligns corpora with pronunciation dictionaries

    See Also
    --------
    :class:`~montreal_forced_aligner.corpus.acoustic_corpus.AcousticCorpusPronunciationMixin`
        For dictionary and corpus parsing parameters
    :class:`~montreal_forced_aligner.alignment.mixins.AlignMixin`
        For alignment parameters
    :class:`~montreal_forced_aligner.abc.FileExporterMixin`
        For file exporting parameters
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.export_output_directory = None

    def word_alignment_arguments(self) -> List[WordAlignmentArguments]:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.alignment.multiprocessing.WordCtmProcessWorker`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.alignment.multiprocessing.WordCtmArguments`]
            Arguments for processing
        """
        return [
            WordAlignmentArguments(
                os.path.join(self.working_log_directory, f"get_word_ctm.{j.name}.log"),
                self.alignment_model_path,
                round(self.frame_shift / 1000, 4),
                self.cleanup_textgrids,
                self.oov_word,
                self.sanitize_function,
                j.current_dictionary_names,
                j.construct_path_dictionary(self.working_directory, "ali", "ark"),
                j.word_boundary_int_files(),
                j.construct_path_dictionary(self.data_directory, "text", "int.scp"),
                {d.name: d.reversed_word_mapping for d in self.dictionary_mapping.values()},
                j.text_scp_data(),
                j.utt2spk_scp_data(),
            )
            for j in self.jobs
        ]

    def phone_alignment_arguments(self) -> List[PhoneAlignmentArguments]:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.alignment.multiprocessing.PhoneCtmProcessWorker`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.alignment.multiprocessing.PhoneCtmArguments`]
            Arguments for processing
        """
        return [
            PhoneAlignmentArguments(
                os.path.join(self.working_log_directory, f"get_phone_ctm.{j.name}.log"),
                self.alignment_model_path,
                round(self.frame_shift / 1000, 4),
                self.position_dependent_phones,
                self.cleanup_textgrids,
                self.optional_silence_phone,
                j.current_dictionary_names,
                j.construct_path_dictionary(self.working_directory, "ali", "ark"),
                j.word_boundary_int_files(),
                j.construct_path_dictionary(self.data_directory, "text", "int.scp"),
                self.reversed_phone_mapping,
            )
            for j in self.jobs
        ]

    def export_textgrid_arguments(self) -> List[ExportTextGridArguments]:
        """
        Generate Job arguments for :class:`~montreal_forced_aligner.alignment.multiprocessing.ExportTextGridProcessWorker`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.alignment.multiprocessing.ExportTextGridArguments`]
            Arguments for processing
        """
        return [
            ExportTextGridArguments(
                os.path.join(self.working_log_directory, f"export_textgrids.{j.name}.log"),
                self.frame_shift,
                self.export_output_directory,
                self.backup_output_directory,
            )
            for j in self.jobs
        ]

    @property
    def backup_output_directory(self) -> Optional[str]:
        """Backup directory if overwriting is not allowed"""
        return None

    def _collect_alignments(self, word_mode=True):
        """
        Process alignment archives to extract word or phone alignments

        Parameters
        ----------
        word_mode: bool
            Flag for collecting word or phone alignments

        See Also
        --------
        :class:`~montreal_forced_aligner.alignment.multiprocessing.WordAlignmentFunction`
            Multiprocessing function for words alignments
        :class:`~montreal_forced_aligner.alignment.multiprocessing.PhoneAlignmentFunction`
            Multiprocessing function for phone alignments
        :meth:`.CorpusAligner.word_alignment_arguments`
            Arguments for word CTMS
        :meth:`.CorpusAligner.phone_alignment_arguments`
            Arguments for phone alignment
        """
        if word_mode:
            self.logger.info("Collecting word alignments from alignment lattices...")
            jobs = self.word_alignment_arguments()  # Word CTM jobs
        else:
            self.logger.info("Collecting phone alignments from alignment lattices...")
            jobs = self.phone_alignment_arguments()  # Phone CTM jobs
        with tqdm.tqdm(total=self.num_utterances) as pbar:
            if self.use_mp:
                manager = mp.Manager()
                error_dict = manager.dict()
                return_queue = manager.Queue()
                stopped = Stopped()
                procs = []
                for i, args in enumerate(jobs):
                    if word_mode:
                        function = WordAlignmentFunction(args)
                    else:
                        function = PhoneAlignmentFunction(args)
                    p = KaldiProcessWorker(i, return_queue, function, error_dict, stopped)
                    procs.append(p)
                    p.start()
                while True:
                    try:
                        utterance, intervals = return_queue.get(timeout=1)
                        if stopped.stop_check():
                            continue
                    except Empty:
                        for proc in procs:
                            if not proc.finished.stop_check():
                                break
                        else:
                            break
                        continue
                    pbar.update(1)
                    if word_mode:
                        self.utterances[utterance].add_word_intervals(intervals)
                    else:
                        self.utterances[utterance].add_phone_intervals(intervals)
                for p in procs:
                    p.join()
                if error_dict:
                    for v in error_dict.values():
                        raise v
            else:
                for args in jobs:
                    if word_mode:
                        function = WordAlignmentFunction(args)
                    else:
                        function = PhoneAlignmentFunction(args)
                    for utterance, intervals in function.run():
                        if word_mode:
                            self.utterances[utterance].add_word_intervals(intervals)
                        else:
                            self.utterances[utterance].add_phone_intervals(intervals)
                        pbar.update(1)

    def collect_word_alignments(self):
        self._collect_alignments(True)

    def collect_phone_alignments(self):
        self._collect_alignments(False)

    def collect_alignments(self):
        if self.alignment_done:
            return
        self.collect_word_alignments()
        self.collect_phone_alignments()
        self.alignment_done = True

    def export_textgrids(self) -> None:
        """
        Exports alignments to TextGrid files

        See Also
        --------
        :class:`~montreal_forced_aligner.alignment.multiprocessing.ExportTextGridProcessWorker`
            Multiprocessing helper function for TextGrid export
        :meth:`.CorpusAligner.export_textgrid_arguments`
            Job method for TextGrid export
        """
        begin = time.time()
        export_directory = self.export_output_directory
        if os.path.exists(export_directory) and not self.overwrite:
            export_directory = self.backup_output_directory
            self.log_debug(f"Not overwriting existing directory, exporting to {export_directory}")

        self.logger.info(f"Exporting TextGrids to {export_directory}...")
        os.makedirs(export_directory, exist_ok=True)
        export_errors = {}
        total_files = len(self.files)
        with tqdm.tqdm(total=total_files) as pbar:
            if self.use_mp:
                manager = mp.Manager()
                textgrid_errors = manager.dict()
                stopped = Stopped()

                finished_processing = Stopped()
                for_write_queue = mp.JoinableQueue()
                export_args = self.export_textgrid_arguments()

                export_procs = []
                for j in self.jobs:
                    export_proc = ExportTextGridProcessWorker(
                        for_write_queue,
                        stopped,
                        finished_processing,
                        textgrid_errors,
                        export_args[j.name],
                    )
                    export_proc.start()
                    export_procs.append(export_proc)
                try:
                    for file in self.files:
                        tiers = file.aligned_data
                        output_path = file.construct_output_path(export_directory)
                        duration = file.duration
                        for_write_queue.put((tiers, output_path, duration))
                        pbar.update(1)
                except Exception:
                    stopped.stop()
                    raise
                finally:
                    finished_processing.stop()

                    for_write_queue.join()
                    for i in range(self.num_jobs):
                        export_procs[i].join()
                    export_errors.update(textgrid_errors)
            else:
                self.log_debug("Not using multiprocessing for TextGrid export")
                for file in self.files:
                    data = file.aligned_data

                    output_path = file.construct_output_path(export_directory)
                    export_textgrid(data, output_path, file.duration, self.frame_shift)
                    pbar.update(1)

        if export_errors:
            self.logger.warning(
                f"There were {len(export_errors)} errors encountered in generating TextGrids. "
                f"Check {os.path.join(export_directory, 'output_errors.txt')} "
                f"for more details"
            )
        output_textgrid_writing_errors(export_directory, export_errors)
        self.logger.info(f"Finished exporting TextGrids to {export_directory}!")
        self.logger.debug(f"Exported TextGrids in a total of {time.time() - begin} seconds")

    def export_files(self, output_directory: str) -> None:
        """
        Export a TextGrid file for every sound file in the dataset

        Parameters
        ----------
        output_directory: str
            Directory to save to
        """
        self.export_output_directory = output_directory
        if self.backup_output_directory is not None and os.path.exists(
            self.backup_output_directory
        ):
            shutil.rmtree(self.backup_output_directory, ignore_errors=True)
        self.export_textgrids()



"""Helper functions for corpus parsing and loading"""
from __future__ import annotations

import os
import shutil
import subprocess
import typing

import soundfile

from montreal_forced_aligner.data import FileExtensions, SoundFileInformation
from montreal_forced_aligner.exceptions import SoxError

SoundFileInfoDict = typing.Dict[str, typing.Union[int, float, str]]

supported_audio_extensions = [".flac", ".ogg", ".aiff", ".mp3"]

__all__ = ["load_text", "find_exts", "get_wav_info"]


def load_text(path: str) -> str:
    """
    Load a text file

    Parameters
    ----------
    path: str
        Text file to load

    Returns
    -------
    str
        Orthographic text of the file
    """
    with open(path, "r", encoding="utf8") as f:
        text = f.read().strip()
    return text


def find_exts(files: typing.List[str]) -> FileExtensions:
    """
    Find and group sound file extensions and transcription file extensions

    Parameters
    ----------
    files: List
        List of filename

    Returns
    -------
    :class:`~montreal_forced_aligner.data.FileExtensions`
        Data class for files found
    """
    exts = FileExtensions([], {}, {}, {}, {})
    for full_filename in files:
        filename, fext = os.path.splitext(full_filename)
        fext = fext.lower()
        if fext == ".wav":
            exts.wav_files[filename] = full_filename
        elif fext == ".lab":
            exts.lab_files[filename] = full_filename
        elif (
            fext == ".txt" and filename not in exts.lab_files
        ):  # .lab files have higher priority than .txt files
            exts.lab_files[filename] = full_filename
        elif fext == ".textgrid":
            exts.textgrid_files[filename] = full_filename
        elif fext in supported_audio_extensions and shutil.which("sox") is not None:
            exts.other_audio_files[filename] = full_filename
        if filename not in exts.identifiers:
            exts.identifiers.append(filename)
    return exts


def get_wav_info(file_path: str) -> SoundFileInformation:
    """
    Get sound file information

    Parameters
    ----------
    file_path: str
        Sound file path

    Returns
    -------
    :class:`~montreal_forced_aligner.data.SoundFileInformation`
        Sound information for format, duration, number of channels, bit depth, and
        sox_string for use in Kaldi feature extraction if necessary
    """
    if file_path.endswith(".mp3"):
        if not shutil.which("soxi"):
            raise SoxError("No sox found")
        sox_proc = subprocess.Popen(
            ["soxi", "-D", file_path], stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True
        )
        stdout, stderr = sox_proc.communicate()
        if stderr.startswith("soxi FAIL formats"):
            raise SoxError("No support for mp3 in sox")
        duration = float(stdout.strip())
        format = "MP3"
        sox_proc = subprocess.Popen(
            ["soxi", "-r", file_path], stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True
        )
        stdout, stderr = sox_proc.communicate()
        sample_rate = int(stdout.strip())
        sox_proc = subprocess.Popen(
            ["soxi", "-c", file_path], stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True
        )
        stdout, stderr = sox_proc.communicate()
        num_channels = int(stdout.strip())
        sox_proc = subprocess.Popen(
            ["soxi", "-p", file_path], stderr=subprocess.PIPE, stdout=subprocess.PIPE, text=True
        )
        stdout, stderr = sox_proc.communicate()
        bit_depth = int(stdout.strip())
        use_sox = True
    else:
        with soundfile.SoundFile(file_path) as inf:
            subtype = inf.subtype
            if subtype == "FLOAT":
                bit_depth = 32
            else:
                bit_depth = int(subtype.split("_")[-1])
            frames = inf.frames
            sample_rate = inf.samplerate
            duration = frames / sample_rate
            num_channels = inf.channels
            format = inf.format
        use_sox = False
        if bit_depth != 16:
            use_sox = True
        if format != "WAV":
            use_sox = True
        if not subtype.startswith("PCM"):
            use_sox = True
    sox_string = ""
    if use_sox:
        sox_string = f"sox {file_path} -t wav -b 16 - |"

    return SoundFileInformation(format, sample_rate, duration, num_channels, bit_depth, sox_string)



"""Class definitions for aligning with pretrained acoustic models"""
from __future__ import annotations

import os
import subprocess
import time
from collections import Counter, defaultdict
from typing import TYPE_CHECKING, Dict, List, NamedTuple, Optional

import yaml

from montreal_forced_aligner.abc import TopLevelMfaWorker
from montreal_forced_aligner.alignment.base import CorpusAligner
from montreal_forced_aligner.exceptions import KaldiProcessingError
from montreal_forced_aligner.helper import align_phones, parse_old_features
from montreal_forced_aligner.models import AcousticModel
from montreal_forced_aligner.utils import log_kaldi_errors, run_mp, run_non_mp, thirdparty_binary

if TYPE_CHECKING:
    from argparse import Namespace

    from montreal_forced_aligner.abc import MetaDict

__all__ = ["PretrainedAligner"]


def generate_pronunciations_func(
    log_path: str,
    dictionaries: List[str],
    text_int_paths: Dict[str, str],
    word_boundary_paths: Dict[str, str],
    ali_paths: Dict[str, str],
    model_path: str,
    pron_paths: Dict[str, str],
):
    """
    Multiprocessing function for generating pronunciations

    See Also
    --------
    :meth:`.DictionaryTrainer.export_lexicons`
        Main function that calls this function in parallel
    :meth:`.DictionaryTrainer.generate_pronunciations_arguments`
        Job method for generating arguments for this function
    :kaldi_src:`linear-to-nbest`
        Kaldi binary this uses

    Parameters
    ----------
    log_path: str
        Path to save log output
    dictionaries: list[str]
        List of dictionary names
    text_int_paths: dict[str, str]
        Dictionary of text int files per dictionary name
    word_boundary_paths: dict[str, str]
        Dictionary of word boundary files per dictionary name
    ali_paths: dict[str, str]
        Dictionary of alignment archives per dictionary name
    model_path: str
        Path to acoustic model file
    pron_paths: dict[str, str]
        Dictionary of pronunciation archives per dictionary name
    """
    with open(log_path, "w", encoding="utf8") as log_file:
        for dict_name in dictionaries:
            text_int_path = text_int_paths[dict_name]
            word_boundary_path = word_boundary_paths[dict_name]
            ali_path = ali_paths[dict_name]
            pron_path = pron_paths[dict_name]

            lin_proc = subprocess.Popen(
                [
                    thirdparty_binary("linear-to-nbest"),
                    f"ark:{ali_path}",
                    f"ark:{text_int_path}",
                    "",
                    "",
                    "ark:-",
                ],
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )
            align_proc = subprocess.Popen(
                [
                    thirdparty_binary("lattice-align-words"),
                    word_boundary_path,
                    model_path,
                    "ark:-",
                    "ark:-",
                ],
                stdin=lin_proc.stdout,
                stdout=subprocess.PIPE,
                stderr=log_file,
                env=os.environ,
            )

            prons_proc = subprocess.Popen(
                [thirdparty_binary("nbest-to-prons"), model_path, "ark:-", pron_path],
                stdin=align_proc.stdout,
                stderr=log_file,
                env=os.environ,
            )
            prons_proc.communicate()


class GeneratePronunciationsArguments(NamedTuple):
    """Arguments for :func:`~montreal_forced_aligner.alignment.pretrained.generate_pronunciations_func`"""

    log_path: str
    dictionaries: List[str]
    text_int_paths: Dict[str, str]
    word_boundary_paths: Dict[str, str]
    ali_paths: Dict[str, str]
    model_path: str
    pron_paths: Dict[str, str]


class PretrainedAligner(CorpusAligner, TopLevelMfaWorker):
    """
    Class for aligning a dataset using a pretrained acoustic model

    Parameters
    ----------
    acoustic_model_path : str
        Path to acoustic model

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.base.CorpusAligner`
        For dictionary and corpus parsing parameters and alignment parameters
    :class:`~montreal_forced_aligner.abc.TopLevelMfaWorker`
        For top-level parameters
    """

    def __init__(
        self,
        acoustic_model_path: str,
        **kwargs,
    ):
        self.acoustic_model = AcousticModel(acoustic_model_path)
        kw = self.acoustic_model.parameters
        kw.update(kwargs)
        super().__init__(**kw)

    @property
    def working_directory(self) -> str:
        """Working directory"""
        return self.workflow_directory

    def setup(self) -> None:
        """Setup for alignment"""
        if self.initialized:
            return
        begin = time.time()
        try:
            os.makedirs(self.working_log_directory, exist_ok=True)
            check = self.check_previous_run()
            if check:
                self.logger.debug(
                    "There were some differences in the current run compared to the last one. "
                    "This may cause issues, run with --clean, if you hit an error."
                )
            self.load_corpus()
            if self.excluded_pronunciation_count:
                self.logger.warning(
                    f"There were {self.excluded_pronunciation_count} pronunciations in the dictionary that"
                    f"were ignored for containing one of {len(self.excluded_phones)} phones not present in the"
                    f"trained acoustic model.  Please run `mfa validate` to get more details."
                )
            self.acoustic_model.validate(self)
            self.acoustic_model.export_model(self.working_directory)
            self.acoustic_model.log_details(self.logger)

        except Exception as e:
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger)
            raise
        self.initialized = True
        self.logger.debug(f"Setup for alignment in {time.time() - begin} seconds")

    @classmethod
    def parse_parameters(
        cls,
        config_path: Optional[str] = None,
        args: Optional[Namespace] = None,
        unknown_args: Optional[List[str]] = None,
    ) -> MetaDict:
        """
        Parse parameters from a config path or command-line arguments

        Parameters
        ----------
        config_path: str
            Config path
        args: :class:`~argparse.Namespace`
            Command-line arguments from argparse
        unknown_args: list[str], optional
            Extra command-line arguments

        Returns
        -------
        dict[str, Any]
            Configuration parameters
        """
        global_params = {}
        if config_path and os.path.exists(config_path):
            with open(config_path, "r", encoding="utf8") as f:
                data = yaml.load(f, Loader=yaml.SafeLoader)
                data = parse_old_features(data)
                for k, v in data.items():
                    if k == "features":
                        global_params.update(v)
                    else:
                        if v is None and k in {
                            "punctuation",
                            "compound_markers",
                            "clitic_markers",
                        }:
                            v = []
                        global_params[k] = v
        global_params.update(cls.parse_args(args, unknown_args))
        return global_params

    @property
    def configuration(self) -> MetaDict:
        """Configuration for aligner"""
        config = super().configuration
        config.update(
            {
                "acoustic_model": self.acoustic_model.name,
            }
        )
        return config

    @property
    def backup_output_directory(self) -> Optional[str]:
        """Backup output directory if overwriting is not allowed"""
        if self.overwrite:
            return None
        return os.path.join(self.working_directory, "textgrids")

    @property
    def workflow_identifier(self) -> str:
        """Aligner identifier"""
        return "pretrained_aligner"

    def evaluate(
        self,
        mapping: Optional[Dict[str, str]] = None,
        output_directory: Optional[str] = None,
    ) -> None:
        """
        Evaluate alignments against a reference directory

        Parameters
        ----------
        reference_directory: str
            Directory containing reference TextGrid alignments
        mapping: dict[str, Union[str, list[str]]], optional
            Mapping between phones that should be considered equal across different phone set types
        output_directory: str, optional
            Directory to save results, if not specified, it will be saved in the log directory
        """
        # Set up
        self.log_info("Evaluating alignments...")
        self.log_debug(f"Mapping: {mapping}")

        score_count = 0
        score_sum = 0
        phone_edit_sum = 0
        phone_length_sum = 0
        if output_directory:
            csv_path = os.path.join(output_directory, "alignment_evaluation.csv")
        else:
            csv_path = os.path.join(self.working_log_directory, "alignment_evaluation.csv")
        with open(csv_path, "w", encoding="utf8") as f:
            f.write(
                "utterance,file,speaker,duration,word_count,oov_count,reference_phone_count,score,phone_error_rate\n"
            )
            for utterance in self.utterances:
                if not utterance.reference_phone_labels:
                    continue
                speaker = utterance.speaker_name
                file = utterance.file_name
                duration = utterance.duration
                reference_phone_count = len(utterance.reference_phone_labels)
                word_count = len(utterance.text.split())
                oov_count = len(utterance.oovs)
                if not utterance.phone_labels:  # couldn't be aligned
                    utterance.alignment_score = None
                    utterance.phone_error_rate = len(utterance.reference_phone_labels)
                    f.write(
                        f"{utterance.name},{file},{speaker},{duration},{word_count},{oov_count},{reference_phone_count},na,{len(utterance.reference_phone_labels)}\n"
                    )

                    continue
                score, phone_error_rate = align_phones(
                    utterance.reference_phone_labels,
                    utterance.phone_labels,
                    self.optional_silence_phone,
                    mapping,
                )
                if score is None:
                    continue
                utterance.alignment_score = score
                utterance.phone_error_rate = phone_error_rate
                f.write(
                    f"{utterance.name},{file},{speaker},{duration},{word_count},{oov_count},{reference_phone_count},{score},{phone_error_rate}\n"
                )
                score_count += 1
                score_sum += score
                phone_edit_sum += int(phone_error_rate * reference_phone_count)
                phone_length_sum += reference_phone_count
        self.logger.info(f"Average overlap score: {score_sum/score_count}")
        self.logger.info(f"Average phone error rate: {phone_edit_sum/phone_length_sum}")

    def align(self) -> None:
        """Run the aligner"""
        self.setup()
        done_path = os.path.join(self.working_directory, "done")
        dirty_path = os.path.join(self.working_directory, "dirty")
        if os.path.exists(done_path):
            self.logger.info("Alignment already done, skipping.")
            return
        try:
            log_dir = os.path.join(self.working_directory, "log")
            os.makedirs(log_dir, exist_ok=True)
            self.compile_train_graphs()

            self.logger.info("Performing first-pass alignment...")
            self.speaker_independent = True
            self.align_utterances()
            self.compile_information()
            if self.uses_speaker_adaptation:
                self.calc_fmllr()

                self.speaker_independent = False
                self.logger.info("Performing second-pass alignment...")
                self.align_utterances()

                self.compile_information()
            self.collect_alignments()
        except Exception as e:
            with open(dirty_path, "w"):
                pass
            if isinstance(e, KaldiProcessingError):
                log_kaldi_errors(e.error_logs, self.logger)
                e.update_log_file(self.logger)
            raise
        with open(done_path, "w"):
            pass


class DictionaryTrainer(PretrainedAligner):
    """
    Aligner for calculating pronunciation probabilities of dictionary entries

    Parameters
    ----------
    calculate_silence_probs: bool
        Flag for whether to calculate silence probabilities, default is False
    min_count: int
        Specifies the minimum count of words to include in derived probabilities,
        affects probabilities of infrequent words more, default is 1

    See Also
    --------
    :class:`~montreal_forced_aligner.alignment.pretrained.PretrainedAligner`
        For dictionary and corpus parsing parameters and alignment parameters
    """

    def __init__(
        self,
        calculate_silence_probs: bool = False,
        min_count: int = 1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.calculate_silence_probs = calculate_silence_probs
        self.min_count = min_count

    def generate_pronunciations_arguments(
        self,
    ) -> List[GeneratePronunciationsArguments]:
        """
        Generate Job arguments for :func:`~montreal_forced_aligner.alignment.pretrained.generate_pronunciations_func`

        Returns
        -------
        list[:class:`~montreal_forced_aligner.alignment.pretrained.GeneratePronunciationsArguments`]
            Arguments for processing
        """
        return [
            GeneratePronunciationsArguments(
                os.path.join(self.working_log_directory, f"generate_pronunciations.{j.name}.log"),
                j.current_dictionary_names,
                j.construct_path_dictionary(self.data_directory, "text", "int.scp"),
                j.word_boundary_int_files(),
                j.construct_path_dictionary(self.working_directory, "ali", "ark"),
                self.model_path,
                j.construct_path_dictionary(self.working_directory, "prons", "scp"),
            )
            for j in self.jobs
        ]

    def export_lexicons(self, output_directory: str) -> None:
        """
        Generate pronunciation probabilities for the dictionary

        Parameters
        ----------
        output_directory: str
            Directory in which to save new dictionaries

        See Also
        --------
        :func:`~montreal_forced_aligner.alignment.pretrained.generate_pronunciations_func`
            Multiprocessing helper function for each job
        :meth:`.DictionaryTrainer.generate_pronunciations_arguments`
            Job method for generating arguments for helper function

        """
        os.makedirs(output_directory, exist_ok=True)
        jobs = self.generate_pronunciations_arguments()
        if self.use_mp:
            run_mp(generate_pronunciations_func, jobs, self.working_log_directory)
        else:
            run_non_mp(generate_pronunciations_func, jobs, self.working_log_directory)
        pron_counts = {}
        utt_mapping = {}
        for j in self.jobs:
            args = jobs[j.name]
            for dict_name, pron_path in args.pron_paths.items():
                if dict_name not in pron_counts:
                    pron_counts[dict_name] = defaultdict(Counter)
                    utt_mapping[dict_name] = {}
                with open(pron_path, "r", encoding="utf8") as f:
                    last_utt = None
                    for line in f:
                        line = line.split()
                        utt = line[0]
                        if utt not in utt_mapping[dict_name]:
                            if last_utt is not None:
                                utt_mapping[dict_name][last_utt].append("</s>")
                            utt_mapping[dict_name][utt] = ["<s>"]
                            last_utt = utt
                        dictionary = self.get_dictionary(self.utterances[utt].speaker_name)
                        word = dictionary.reversed_word_mapping[int(line[3])]
                        if word == "<eps>":
                            utt_mapping[dict_name][utt].append(word)
                        else:
                            pron = tuple(
                                dictionary.reversed_phone_mapping[int(x)].split("_")[0]
                                for x in line[4:]
                            )
                            pron_string = " ".join(pron)
                            utt_mapping[dict_name][utt].append(word + " " + pron_string)
                            pron_counts[dict_name][word][pron] += 1
        for dict_name, dictionary in self.dictionary_mapping.items():
            counts = pron_counts[dict_name]
            mapping = utt_mapping[dict_name]
            if self.calculate_silence_probs:
                sil_before_counts = Counter()
                nonsil_before_counts = Counter()
                sil_after_counts = Counter()
                nonsil_after_counts = Counter()
                sils = ["<s>", "</s>", "<eps>"]
                for v in mapping.values():
                    for i, w in enumerate(v):
                        if w in sils:
                            continue
                        prev_w = v[i - 1]
                        next_w = v[i + 1]
                        if prev_w in sils:
                            sil_before_counts[w] += 1
                        else:
                            nonsil_before_counts[w] += 1
                        if next_w in sils:
                            sil_after_counts[w] += 1
                        else:
                            nonsil_after_counts[w] += 1

            dictionary.pronunciation_probabilities = True
            for word, prons in dictionary.actual_words.items():
                if word not in counts:
                    for p in prons:
                        p.probability = 1
                else:
                    total = 0
                    best_pron = 0
                    best_count = 0
                    for p in prons:
                        p.probability = self.min_count
                        if p.pronunciation in counts[word]:
                            p.probability += counts[word][p.pronunciation]
                        total += p.probability
                        if p.probability > best_count:
                            best_pron = p.pronunciation
                            best_count = p.probability
                    for p in prons:
                        if p.pronunciation == best_pron:
                            p.probability = 1
                        else:
                            p.probability /= total
                    dictionary.words[word] = prons
            output_path = os.path.join(output_directory, dict_name + ".txt")
            dictionary.export_lexicon(output_path, probability=True)



"""Class definitions for corpora"""
from __future__ import annotations

import json
import os
import random
import time
from abc import ABCMeta, abstractmethod
from collections import Counter
from typing import Dict, List, Optional, Union

import jsonlines
import yaml

from montreal_forced_aligner.abc import MfaWorker, TemporaryDirectoryMixin
from montreal_forced_aligner.corpus.classes import (
    File,
    FileCollection,
    Speaker,
    SpeakerCollection,
    Utterance,
    UtteranceCollection,
)
from montreal_forced_aligner.corpus.multiprocessing import Job
from montreal_forced_aligner.data import CtmInterval, SoundFileInformation
from montreal_forced_aligner.exceptions import CorpusError
from montreal_forced_aligner.helper import jsonl_encoder, output_mapping
from montreal_forced_aligner.utils import Stopped

__all__ = ["CorpusMixin"]


class CorpusMixin(MfaWorker, TemporaryDirectoryMixin, metaclass=ABCMeta):
    """
    Mixin class for processing corpora

    Notes
    -----
    Using characters in files to specify speakers is generally finicky and leads to errors, so I would not
    recommend using it.  Additionally, consider it deprecated and could be removed in future versions

    Parameters
    ----------
    corpus_directory: str
        Path to corpus
    speaker_characters: int or str, optional
        Number of characters in the file name to specify the speaker
    ignore_speakers: bool
        Flag for whether to discard any parsed speaker information during top-level worker's processing

    See Also
    --------
    :class:`~montreal_forced_aligner.abc.MfaWorker`
        For MFA processing parameters
    :class:`~montreal_forced_aligner.abc.TemporaryDirectoryMixin`
        For temporary directory parameters

    Attributes
    ----------
    speakers: :class:`~montreal_forced_aligner.corpus.classes.SpeakerCollection`
        Dictionary of speakers in the corpus
    files: :class:`~montreal_forced_aligner.corpus.classes.FileCollection`
        Dictionary of files in the corpus
    utterances: :class:`~montreal_forced_aligner.corpus.classes.UtteranceCollection`
        Dictionary of utterances in the corpus
    jobs: list[Job]
        List of jobs for processing the corpus and splitting speakers
    word_counts: Counter
        Counts of words in the corpus
    stopped: Stopped
        Stop check for loading the corpus
    decode_error_files: list[str]
        List of text files that could not be loaded with utf8
    textgrid_read_errors: list[str]
        List of TextGrid files that had an error in loading
    """

    def __init__(
        self,
        corpus_directory: str,
        speaker_characters: Union[int, str] = 0,
        ignore_speakers: bool = False,
        **kwargs,
    ):
        if not os.path.exists(corpus_directory):
            raise CorpusError(f"The directory '{corpus_directory}' does not exist.")
        if not os.path.isdir(corpus_directory):
            raise CorpusError(
                f"The specified path for the corpus ({corpus_directory}) is not a directory."
            )
        self.speakers = SpeakerCollection()
        self.files = FileCollection()
        self.utterances = UtteranceCollection()
        self.corpus_directory = corpus_directory
        self.speaker_characters = speaker_characters
        self.ignore_speakers = ignore_speakers
        self.word_counts = Counter()
        self.stopped = Stopped()
        self.decode_error_files = []
        self.textgrid_read_errors = {}
        self.jobs: List[Job] = []
        super().__init__(**kwargs)

    def _initialize_from_json(self, data):
        pass

    @property
    def corpus_meta(self):
        return {}

    @property
    def features_directory(self) -> str:
        """Feature directory of the corpus"""
        return os.path.join(self.corpus_output_directory, "features")

    @property
    def features_log_directory(self) -> str:
        """Feature log directory"""
        return os.path.join(self.split_directory, "log")

    @property
    def split_directory(self) -> str:
        """Directory used to store information split by job"""
        return os.path.join(self.corpus_output_directory, f"split{self.num_jobs}")

    def write_corpus_information(self) -> None:
        """
        Output information to the temporary directory for later loading
        """
        os.makedirs(self.split_directory, exist_ok=True)
        self._write_speakers()
        self._write_files()
        self._write_utterances()
        self._write_spk2utt()
        self._write_corpus_info()

    def _write_spk2utt(self):
        """Write spk2utt scp file for Kaldi"""
        data = {
            speaker.name: sorted(u.name for u in speaker.utterances) for speaker in self.speakers
        }
        output_mapping(data, os.path.join(self.corpus_output_directory, "spk2utt.scp"))

    def write_utt2spk(self):
        """Write utt2spk scp file for Kaldi"""
        data = {u.name: u.speaker.name for u in self.utterances}
        output_mapping(data, os.path.join(self.corpus_output_directory, "utt2spk.scp"))

    def _write_corpus_info(self):
        """Write speaker information for speeding up future runs"""
        with open(
            os.path.join(self.corpus_output_directory, "corpus.json"), "w", encoding="utf8"
        ) as f:
            json.dump(self.corpus_meta, f)

    def _write_speakers(self):
        """Write speaker information for speeding up future runs"""
        with open(
            os.path.join(self.corpus_output_directory, "speakers.jsonl"), "w", encoding="utf8"
        ) as f:
            writer = jsonlines.Writer(f, dumps=jsonl_encoder)
            for speaker in self.speakers:
                writer.write(speaker.meta)

    def _write_files(self):
        """Write file information for speeding up future runs"""
        with open(
            os.path.join(self.corpus_output_directory, "files.jsonl"), "w", encoding="utf8"
        ) as f:
            writer = jsonlines.Writer(f, dumps=jsonl_encoder)
            for file in self.files:
                writer.write(file.meta)

    def _write_utterances(self):
        """Write utterance information for speeding up future runs"""
        with open(
            os.path.join(self.corpus_output_directory, "utterances.jsonl"), "w", encoding="utf8"
        ) as f:
            writer = jsonlines.Writer(f, dumps=jsonl_encoder)
            for utterance in self.utterances:
                writer.write(utterance.meta)

    def create_corpus_split(self) -> None:
        """Create split directory and output information from Jobs"""
        split_dir = self.split_directory
        os.makedirs(os.path.join(split_dir, "log"), exist_ok=True)
        for job in self.jobs:
            job.output_to_directory(split_dir)

    @property
    def file_speaker_mapping(self) -> Dict[str, List[str]]:
        """Speaker ordering for each file"""
        return {file.name: file.speaker_ordering for file in self.files}

    def get_word_frequency(self) -> Dict[str, float]:
        """
        Calculate the relative word frequency across all the texts in the corpus

        Returns
        -------
        dict[str, float]
            Dictionary of words and their relative frequencies
        """
        word_counts = Counter()
        for u in self.utterances:
            text = u.text
            speaker = u.speaker
            d = speaker.dictionary
            new_text = []
            text = text.split()
            for t in text:

                lookup = d.split_clitics(t)
                if lookup is None:
                    continue
                new_text.extend(x for x in lookup if x != "")
            word_counts.update(new_text)
        return {k: v / sum(word_counts.values()) for k, v in word_counts.items()}

    @property
    def corpus_word_set(self) -> List[str]:
        """Set of words used in the corpus"""
        return sorted(self.word_counts)

    def add_utterance(self, utterance: Utterance) -> None:
        """
        Add an utterance to the corpus

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to add
        """
        self.utterances.add_utterance(utterance)
        if utterance.speaker not in self.speakers:
            self.speakers.add_speaker(utterance.speaker)
        speaker = self.speakers[utterance.speaker.name]
        speaker.add_utterance(utterance)
        utterance.speaker = speaker
        if utterance.file not in self.files:
            self.files.add_file(utterance.file)
        file = self.files[utterance.file.name]
        utterance.file = file
        file.add_utterance(utterance)
        utterance.file = file

    def delete_utterance(self, utterance: Union[str, Utterance]) -> None:
        """
        Delete an utterance from the corpus

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to delete
        """
        if isinstance(utterance, str):
            utterance = self.utterances[utterance]
        speaker = self.speakers[utterance.speaker.name]
        file = self.files[utterance.file.name]
        speaker.delete_utterance(utterance)
        file.delete_utterance(utterance)
        del self.utterances[utterance.name]

    def initialize_jobs(self) -> None:
        """
        Initialize the corpus's Jobs
        """
        self.log_info("Setting up training data...")
        if len(self.speakers) < self.num_jobs:
            self.num_jobs = len(self.speakers)
        self.jobs = [Job(i) for i in range(self.num_jobs)]
        job_ind = 0
        for s in sorted(self.speakers):
            self.jobs[job_ind].add_speaker(s)
            job_ind += 1
            if job_ind == self.num_jobs:
                job_ind = 0

    def add_file(self, file: File) -> None:
        """
        Add a file to the corpus

        Parameters
        ----------
        file: :class:`~montreal_forced_aligner.corpus.classes.File`
            File to be added
        """
        self.files.add_file(file)
        for i, speaker in enumerate(file.speaker_ordering):
            if speaker.name not in self.speakers:
                self.speakers.add_speaker(speaker)
            else:
                file.speaker_ordering[i] = self.speakers[speaker.name]
        for u in file.utterances:
            self.add_utterance(u)
            if u.text:
                self.word_counts.update(u.text.split())
            if u.normalized_text:
                self.word_counts.update(u.normalized_text)

    @property
    def data_source_identifier(self) -> str:
        """Corpus name"""
        return os.path.basename(self.corpus_directory)

    def create_subset(self, subset: int) -> None:
        """
        Create a subset of utterances to use for training

        Parameters
        ----------
        subset: int
            Number of utterances to include in subset
        """
        self.log_debug(f"Creating subset_{subset} directory...")
        subset_directory = os.path.join(self.corpus_output_directory, f"subset_{subset}")

        larger_subset_num = subset * 10
        if larger_subset_num < len(self.utterances):
            # Get all shorter utterances that are not one word long
            utts = sorted(
                (utt for utt in self.utterances if utt.text and " " in utt.text),
                key=lambda x: x.duration,
            )
            larger_subset = utts[:larger_subset_num]
        else:
            larger_subset = sorted(self.utterances)
        random.seed(1234)  # make it deterministic sampling
        subset_utts = UtteranceCollection()
        subset_utts.update(random.sample(larger_subset, subset))
        log_dir = os.path.join(subset_directory, "log")
        os.makedirs(log_dir, exist_ok=True)

        for j in self.jobs:
            j.set_subset(subset_utts)
            j.output_to_directory(subset_directory)

    @property
    def num_utterances(self) -> int:
        """Number of utterances in the corpus"""
        return len(self.utterances)

    @property
    def num_speakers(self) -> int:
        """Number of speakers in the corpus"""
        return len(self.speakers)

    def subset_directory(self, subset: Optional[int]) -> str:
        """
        Construct a subset directory for the corpus

        Parameters
        ----------
        subset: int, optional
            Number of utterances to include, if larger than the total number of utterance or not specified, the
            split_directory is returned

        Returns
        -------
        str
            Path to subset directory
        """
        if subset is None or subset >= len(self.utterances) or subset <= 0:
            for j in self.jobs:
                j.set_subset(None)
            return self.split_directory
        directory = os.path.join(self.corpus_output_directory, f"subset_{subset}")
        if not os.path.exists(directory):
            self.create_subset(subset)
        return directory

    def _load_corpus(self) -> None:
        """
        Load the corpus
        """
        self.log_info("Setting up corpus information...")
        loaded = self._load_corpus_from_temp()
        if not loaded:
            self.log_debug("Could not load from temp")
            self.log_info("Loading corpus from source files...")
            if self.use_mp:

                self._load_corpus_from_source_mp()
            else:
                self._load_corpus_from_source()
        else:
            self.log_debug("Successfully loaded from temporary files")
        if not self.files:
            raise CorpusError(
                "There were no files found for this corpus. Please validate the corpus."
            )
        num_speakers = len(self.speakers)
        if not num_speakers:
            raise CorpusError(
                "There were no sound files found of the appropriate format. Please double check the corpus path "
                "and/or run the validation utility (mfa validate)."
            )
        average_utterances = sum(len(x.utterances) for x in self.speakers) / num_speakers
        self.log_info(
            f"Number of speakers in corpus: {num_speakers}, "
            f"average number of utterances per speaker: {average_utterances}"
        )

    def _load_corpus_from_temp(self) -> bool:
        """
        Load a corpus from saved data in the temporary directory

        Returns
        -------
        bool
            Whether loading from temporary files was successful
        """
        begin_time = time.time()
        if not os.path.exists(self.corpus_output_directory):
            return False
        for f in os.listdir(self.corpus_output_directory):
            if f.startswith("split"):
                old_num_jobs = int(f.replace("split", ""))
                if old_num_jobs != self.num_jobs:
                    self.log_info(
                        f"Found old run with {old_num_jobs} rather than the current {self.num_jobs}, "
                        f"setting to {old_num_jobs}.  If you would like to use {self.num_jobs}, re-run the command "
                        f"with --clean."
                    )
                    self.num_jobs = old_num_jobs
        format = "jsonl"
        corpus_path = os.path.join(self.corpus_output_directory, "corpus.json")
        speakers_path = os.path.join(self.corpus_output_directory, "speakers.jsonl")
        files_path = os.path.join(self.corpus_output_directory, "files.jsonl")
        utterances_path = os.path.join(self.corpus_output_directory, "utterances.jsonl")
        if not os.path.exists(speakers_path):
            format = "yaml"
            speakers_path = os.path.join(self.corpus_output_directory, "speakers.yaml")
            files_path = os.path.join(self.corpus_output_directory, "files.yaml")
            utterances_path = os.path.join(self.corpus_output_directory, "utterances.yaml")

        if not os.path.exists(speakers_path):
            self.log_debug(f"Could not find {speakers_path}, cannot load from temp")
            return False
        if not os.path.exists(files_path):
            self.log_debug(f"Could not find {files_path}, cannot load from temp")
            return False
        if not os.path.exists(utterances_path):
            self.log_debug(f"Could not find {utterances_path}, cannot load from temp")
            return False
        self.log_debug("Loading from temporary files...")

        if os.path.exists(corpus_path):
            with open(corpus_path, "r", encoding="utf8") as f:
                data = json.load(f)
                self._initialize_from_json(data)

        with open(speakers_path, "r", encoding="utf8") as f:
            if format == "jsonl":
                speaker_data = jsonlines.Reader(f)
            else:
                speaker_data = yaml.safe_load(f)

            for entry in speaker_data:
                self.speakers.add_speaker(Speaker(entry["name"]))
                self.speakers[entry["name"]].cmvn = entry["cmvn"]

        with open(files_path, "r", encoding="utf8") as f:
            if format == "jsonl":
                files_data = jsonlines.Reader(f)
            else:
                files_data = yaml.safe_load(f)
            for entry in files_data:
                file = File(
                    name=entry["name"],
                    wav_path=entry["wav_path"],
                    text_path=entry["text_path"],
                    relative_path=entry["relative_path"],
                )
                self.files.add_file(file)
                self.files[file.name].speaker_ordering = [
                    self.speakers[x] for x in entry["speaker_ordering"]
                ]
                self.files[file.name].wav_info = SoundFileInformation(**entry["wav_info"])

        with open(utterances_path, "r", encoding="utf8") as f:
            if format == "jsonl":
                utterances_data = jsonlines.Reader(f)
            else:
                utterances_data = yaml.safe_load(f)
            for entry in utterances_data:
                s = self.speakers[entry["speaker"]]
                f = self.files[entry["file"]]
                u = Utterance(
                    s,
                    f,
                    begin=entry["begin"],
                    end=entry["end"],
                    channel=entry["channel"],
                    text=entry["text"],
                )
                u.oovs = set(entry["oovs"])
                u.normalized_text = entry["normalized_text"]
                if u.text:
                    self.word_counts.update(u.text.split())
                if u.normalized_text:
                    self.word_counts.update(u.normalized_text)
                u.word_error_rate = entry.get("word_error_rate", None)
                u.character_error_rate = entry.get("character_error_rate", None)
                u.transcription_text = entry.get("transcription_text", None)
                u.phone_error_rate = entry.get("phone_error_rate", None)
                u.alignment_score = entry.get("alignment_score", None)
                u.alignment_log_likelihood = entry.get("alignment_log_likelihood", None)
                u.reference_phone_labels = [
                    CtmInterval(**x) for x in entry.get("reference_phone_labels", [])
                ]

                phone_labels = entry.get("phone_labels", None)
                if phone_labels:
                    u.phone_labels = [CtmInterval(**x) for x in phone_labels]
                word_labels = entry.get("word_labels", None)
                if word_labels:
                    u.word_labels = [CtmInterval(**x) for x in word_labels]
                u.features = entry.get("features", None)
                u.ignored = entry.get("ignored", False)
                self.add_utterance(u)

        self.log_debug(
            f"Loaded from corpus_data temp directory in {time.time() - begin_time} seconds"
        )
        return True

    @property
    def base_data_directory(self) -> str:
        """Corpus data directory"""
        return self.corpus_output_directory

    @property
    def data_directory(self) -> str:
        """Corpus data directory"""
        return self.split_directory

    @abstractmethod
    def _load_corpus_from_source_mp(self) -> None:
        """Abstract method for loading a corpus with multiprocessing"""
        ...

    @abstractmethod
    def _load_corpus_from_source(self) -> None:
        """Abstract method for loading a corpus without multiprocessing"""
        ...



"""Class definitions for Speakers, Files, Utterances and Jobs"""
from __future__ import annotations

import os
import re
import sys
import traceback
from collections import Counter
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Dict,
    Generator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import librosa
import numpy as np
from praatio import textgrid
from praatio.utilities.constants import Interval

from montreal_forced_aligner.abc import MfaCorpusClass
from montreal_forced_aligner.corpus.helper import get_wav_info, load_text
from montreal_forced_aligner.data import (
    CtmInterval,
    FileData,
    SoundFileInformation,
    SoundFileType,
    TextFileType,
    UtteranceData,
)
from montreal_forced_aligner.dictionary.multispeaker import MultispeakerSanitizationFunction
from montreal_forced_aligner.exceptions import CorpusError, TextGridParseError, TextParseError

if TYPE_CHECKING:
    from montreal_forced_aligner.dictionary.pronunciation import PronunciationDictionaryMixin


__all__ = ["File", "Speaker", "Utterance"]


class Speaker(MfaCorpusClass):
    """
    Class representing information about a speaker

    Parameters
    ----------
    name: str
        Identifier of the speaker

    Attributes
    ----------
    utterances: :class:`~montreal_forced_aligner.corpus.classes.UtteranceCollection`
        Utterances that the speaker is associated with
    cmvn: str, optional
        String pointing to any CMVN that has been calculated for this speaker
    dictionary: :class:`~montreal_forced_aligner.dictionary.PronunciationDictionary`, optional
        Pronunciation dictionary that the speaker is associated with
    """

    def __init__(self, name):
        self._name = name
        self.utterances = UtteranceCollection()
        self.cmvn = None
        self.dictionary: Optional[PronunciationDictionaryMixin] = None
        self.dictionary_name: Optional[str] = None
        self.word_counts = Counter()

    @property
    def name(self) -> str:
        """Speaker name"""
        return self._name

    def __str__(self) -> str:
        """Return Speaker's name"""
        return self.name

    def __eq__(self, other: Union[Speaker, str]) -> bool:
        """Check if a Speaker is equal to another Speaker"""
        if isinstance(other, Speaker):
            return other.name == self.name
        if isinstance(other, str):
            return self.name == other
        raise TypeError("Speakers can only be compared to other speakers and strings.")

    def __lt__(self, other: Union[Speaker, str]) -> bool:
        """Check if a Speaker is less than another Speaker"""
        if isinstance(other, Speaker):
            return other.name < self.name
        if isinstance(other, str):
            return self.name < other
        raise TypeError("Speakers can only be compared to other speakers and strings.")

    def __gt__(self, other: Union[Speaker, str]) -> bool:
        """Check if a Speaker is greater than another Speaker"""
        if isinstance(other, Speaker):
            return other.name > self.name
        if isinstance(other, str):
            return self.name > other
        raise TypeError("Speakers can only be compared to other speakers and strings.")

    def __hash__(self) -> hash:
        """Get the hash of the speaker"""
        return hash(self.name)

    @property
    def num_utterances(self) -> int:
        """Get the number of utterances for the speaker"""
        return len(self.utterances)

    def add_utterance(self, utterance: Utterance) -> None:
        """
        Associate an utterance with a speaker

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to be added
        """
        self.utterances.add_utterance(utterance)

    def delete_utterance(self, utterance: Utterance) -> None:
        """
        Delete an utterance associated with a speaker

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to be deleted
        """
        identifier = utterance.name
        del self.utterances[identifier]

    def set_dictionary(self, dictionary: PronunciationDictionaryMixin) -> None:
        """
        Set the dictionary for the speaker

        Parameters
        ----------
        dictionary: :class:`~montreal_forced_aligner.dictionary.PronunciationDictionary`
            Pronunciation dictionary to associate with the speaker
        """
        self.dictionary = dictionary
        self.dictionary_name = dictionary.name

    @property
    def meta(self) -> Dict[str, str]:
        """Metadata for the speaker"""
        data = {
            "name": self.name,
            "cmvn": self.cmvn,
        }
        if self.dictionary is not None:
            data["dictionary"] = self.dictionary.name
        return data


class File(MfaCorpusClass):
    """
    File class for representing metadata and associations of Files

    Parameters
    ----------
    wav_path: str, optional
        Sound file path
    text_path: str, optional
        Transcription file path
    relative_path: str, optional
        Relative path to the corpus root

    Attributes
    ----------
    utterances: :class:`~montreal_forced_aligner.corpus.classes.UtteranceCollection`
        Utterances in the file
    speaker_ordering: list[Speaker]
        Ordering of speakers in the transcription file
    wav_info: :class:`~montreal_forced_aligner.data.SoundFileInformation`
        Information about sound file
    waveform: numpy.array
        Audio samples
    aligned: bool
        Flag for whether a file has alignments

    Raises
    ------
    :class:`~montreal_forced_aligner.exceptions.CorpusError`
        If both wav_path and text_path are None
    """

    textgrid_regex = re.compile(r"\.textgrid$", flags=re.IGNORECASE)
    wav_regex = re.compile(r"\.wav$", flags=re.IGNORECASE)

    def __init__(
        self,
        wav_path: Optional[str] = None,
        text_path: Optional[str] = None,
        relative_path: Optional[str] = None,
        name: Optional[str] = None,
    ):
        self.wav_path = wav_path
        self.text_path = text_path
        wav_check = self.wav_path is not None
        text_check = self.text_path is not None
        self._name = name
        if not self._name:
            if wav_check:
                self._name = os.path.splitext(os.path.basename(self.wav_path))[0]
            elif text_check:
                self._name = os.path.splitext(os.path.basename(self.text_path))[0]
            else:
                raise CorpusError("File objects must have either a wav_path or text_path")
        self.relative_path = relative_path
        self.wav_info: Optional[SoundFileInformation] = None
        self.waveform = None
        self.speaker_ordering: List[Speaker] = []
        self.utterances = UtteranceCollection()
        self.aligned = False
        if text_check:
            if self.text_path.lower().endswith(".textgrid"):
                self.text_type = TextFileType.TEXTGRID
            else:
                self.text_type = TextFileType.LAB
        else:
            self.text_type = TextFileType.NONE
        if wav_check:

            if self.wav_path.lower().endswith(".wav"):
                self.sound_type = SoundFileType.WAV
            else:
                self.sound_type = SoundFileType.SOX
        else:
            self.sound_type = SoundFileType.NONE

    @property
    def multiprocessing_data(self) -> FileData:
        """
        Data object for the file
        """
        return FileData(
            self._name,
            self.wav_path,
            self.text_path,
            self.relative_path,
            self.wav_info,
            [s.name for s in self.speaker_ordering],
            [u.multiprocessing_data for u in self.utterances],
        )

    @classmethod
    def load_from_mp_data(cls, file_data: FileData) -> File:
        """
        Construct a File from a multiprocessing file data class

        Parameters
        ----------
        file_data: :class:`~montreal_forced_aligner.data.FileData`
            Data for the loaded file

        Returns
        -------
        :class:`~montreal_forced_aligner.corpus.classes.File`
            Loaded file
        """
        file = File(
            file_data.wav_path,
            file_data.text_path,
            relative_path=file_data.relative_path,
            name=file_data.name,
        )
        file.wav_info = file_data.wav_info
        for s in file_data.speaker_ordering:
            file.add_speaker(Speaker(s))
        for u in file_data.utterances:
            u = Utterance.load_from_mp_data(u, file)
            file.utterances.add_utterance(u)
        return file

    @classmethod
    def parse_file(
        cls,
        file_name: str,
        wav_path: Optional[str],
        text_path: Optional[str],
        relative_path: str,
        speaker_characters: Union[int, str],
        sanitize_function: Optional[MultispeakerSanitizationFunction] = None,
    ):
        """
        Parse a collection of sound file and transcription file into a File

        Parameters
        ----------
        file_name: str
            File identifier
        wav_path: str
            Full sound file path
        text_path: str
            Full transcription path
        relative_path: str
            Relative path from the corpus directory root
        speaker_characters: int, optional
            Number of characters in the file name to specify the speaker
        sanitize_function: Callable, optional
            Function to sanitize words and strip punctuation

        Returns
        -------
        :class:`~montreal_forced_aligner.corpus.classes.File`
            Parsed file
        """
        file = File(wav_path, text_path, relative_path=relative_path, name=file_name)
        if file.has_sound_file:
            root = os.path.dirname(wav_path)
            file.wav_info = get_wav_info(wav_path)
        else:
            root = os.path.dirname(text_path)
        if not speaker_characters:
            speaker_name = os.path.basename(root)
        elif isinstance(speaker_characters, int):
            speaker_name = file_name[:speaker_characters]
        elif speaker_characters == "prosodylab":
            speaker_name = file_name.split("_")[1]
        else:
            speaker_name = file_name
        root_speaker = None
        if speaker_characters or file.text_type != TextFileType.TEXTGRID:
            root_speaker = Speaker(speaker_name)
        file.load_text(
            root_speaker=root_speaker,
            sanitize_function=sanitize_function,
        )
        return file

    def __eq__(self, other: Union[File, str]) -> bool:
        """Check if a File is equal to another File"""
        if isinstance(other, File):
            return other.name == self.name
        if isinstance(other, str):
            return self.name == other
        raise TypeError("Files can only be compared to other files and strings.")

    def __lt__(self, other: Union[File, str]) -> bool:
        """Check if a File is less than another File"""
        if isinstance(other, File):
            return other.name < self.name
        if isinstance(other, str):
            return self.name < other
        raise TypeError("Files can only be compared to other files and strings.")

    def __gt__(self, other: Union[File, str]) -> bool:
        """Check if a File is greater than another File"""
        if isinstance(other, File):
            return other.name > self.name
        if isinstance(other, str):
            return self.name > other
        raise TypeError("Files can only be compared to other files and strings.")

    def __hash__(self) -> hash:
        """Get the hash of the file"""
        return hash(self.name)

    @property
    def name(self) -> str:
        """Name of the file"""
        name = self._name
        if self.relative_path:
            prefix = self.relative_path.replace("/", "").replace("\\", "")
            if not name.startswith(prefix):
                name = f"{prefix}_{name}"
        return name

    @property
    def is_fully_aligned(self) -> bool:
        """
        Check if all utterances have been aligned
        """
        for u in self.utterances:
            if u.ignored:
                continue
            if u.word_labels is None:
                return False
            if u.phone_labels is None:
                return False
        return True

    def __repr__(self) -> str:
        """Representation of File objects"""
        return f'<File {self.name} Sound path="{self.wav_path}" Text path="{self.text_path}">'

    def save(
        self,
        output_directory: Optional[str] = None,
        text_type: Optional[TextFileType] = None,
        save_transcription: bool = False,
    ) -> None:
        """
        Output File to TextGrid or lab.  If ``text_type`` is not specified, the original file type will be used,
        but if there was no text file for the file, it will guess lab format if there is only one utterance, otherwise
        it will output a TextGrid file.

        Parameters
        ----------
        output_directory: str, optional
            Directory to output file, if None, then it will overwrite the original file
        backup_output_directory: str, optional
            If specified, then it will check whether it would overwrite an existing file and
            instead use this directory
        text_type: TextFileType, optional
            Text type to save as, if not provided, it will use either the original file type or guess the file type
        save_transcription: bool
            Flag for whether the hypothesized transcription text should be saved instead of the default text
        """
        utterance_count = len(self.utterances)
        if text_type is None:
            text_type = self.text_type
            if text_type == TextFileType.NONE:
                if utterance_count == 1:
                    text_type = TextFileType.LAB
                else:
                    text_type = TextFileType.TEXTGRID
        if text_type == TextFileType.LAB:
            if utterance_count == 0 and os.path.exists(self.text_path) and not save_transcription:
                os.remove(self.text_path)
                return
            elif utterance_count == 0:
                return
            output_path = self.construct_output_path(output_directory, enforce_lab=True)
            with open(output_path, "w", encoding="utf8") as f:
                for u in self.utterances:
                    if save_transcription:
                        f.write(u.transcription_text if u.transcription_text else "")
                    else:
                        f.write(u.text)
            return
        elif text_type == TextFileType.TEXTGRID:
            output_path = self.construct_output_path(output_directory)
            max_time = self.duration
            tiers = {}
            for speaker in self.speaker_ordering:
                if speaker is None:
                    tiers["speech"] = textgrid.IntervalTier("speech", [], minT=0, maxT=max_time)
                else:
                    tiers[speaker] = textgrid.IntervalTier(speaker.name, [], minT=0, maxT=max_time)

            tg = textgrid.Textgrid()
            tg.maxTimestamp = max_time
            for utterance in self.utterances:

                if utterance.speaker is None:
                    speaker = "speech"
                else:
                    speaker = utterance.speaker
                if not self.aligned:

                    if save_transcription:
                        tiers[speaker].entryList.append(
                            Interval(
                                start=utterance.begin,
                                end=utterance.end,
                                label=utterance.transcription_text
                                if utterance.transcription_text
                                else "",
                            )
                        )
                    else:
                        tiers[speaker].entryList.append(
                            Interval(
                                start=utterance.begin, end=utterance.end, label=utterance.text
                            )
                        )
            for t in tiers.values():
                tg.addTier(t)
            tg.save(output_path, includeBlankSpaces=True, format="long_textgrid")

    @property
    def meta(self) -> Dict[str, Any]:
        """Metadata for the File"""
        return {
            "wav_path": self.wav_path,
            "text_path": self.text_path,
            "name": self._name,
            "relative_path": self.relative_path,
            "wav_info": self.wav_info.meta,
            "speaker_ordering": [x.name for x in self.speaker_ordering],
        }

    @property
    def has_sound_file(self) -> bool:
        """Flag for whether the File has a sound file"""
        return self.sound_type != SoundFileType.NONE

    @property
    def has_text_file(self) -> bool:
        """Flag for whether the File has a text file"""
        return self.text_type != TextFileType.NONE

    @property
    def aligned_data(self) -> Dict[str, Dict[str, List[CtmInterval]]]:
        """
        Word and phone alignments for the file

        Returns
        -------
        dict[str, dict[str, list[CtmInterval]]]
            Dictionary of word and phone intervals for each speaker in the file
        """
        data = {}
        for s in self.speaker_ordering:
            if s.name not in data:
                data[s.name] = {"words": [], "phones": []}
        for u in self.utterances:
            if u.word_labels is None:
                continue
            data[u.speaker_name]["words"].extend(u.word_labels)
            data[u.speaker_name]["phones"].extend(u.phone_labels)
        return data

    def clean_up(self) -> None:
        """
        Recombine words that were split up as part of initial text processing
        """
        for u in self.utterances:
            if not u.word_labels:
                continue
            cur_ind = 0
            actual_labels = []
            dictionary = u.speaker.dictionary
            for word in u.text.split():
                splits = dictionary.lookup(word)
                b = 1000000
                e = -1
                for w in splits:
                    cur = u.word_labels[cur_ind]
                    if w == cur.label or cur.label == dictionary.oov_word:
                        if cur.begin < b:
                            b = cur.begin
                        if cur.end > e:
                            e = cur.end
                    cur_ind += 1
                lab = CtmInterval(b, e, word, u.name)
                actual_labels.append(lab)
            u.word_labels = actual_labels
            u.phone_labels = [
                x for x in u.phone_labels if x.label != dictionary.optional_silence_phone
            ]

    def construct_output_path(
        self,
        output_directory: Optional[str] = None,
        enforce_lab: bool = False,
    ) -> str:
        """
        Construct the output path for the File

        Parameters
        ----------
        output_directory: str, optional
            Directory to output to, if None, it will overwrite the original file
        enforce_lab: bool
            Flag for whether to enforce generating a lab file over a TextGrid

        Returns
        -------
        str
            Output path
        """
        if enforce_lab:
            extension = ".lab"
        else:
            extension = ".TextGrid"
        if output_directory is None:
            if self.text_path is None:
                return os.path.splitext(self.wav_path)[0] + extension
            return self.text_path
        if self.relative_path:
            relative = os.path.join(output_directory, self.relative_path)
        else:
            relative = output_directory
        tg_path = os.path.join(relative, self._name + extension)
        os.makedirs(os.path.dirname(tg_path), exist_ok=True)
        return tg_path

    def load_text(
        self,
        root_speaker: Optional[Speaker] = None,
        sanitize_function: Optional[MultispeakerSanitizationFunction] = None,
    ) -> None:
        """
        Load the transcription text from the text_file of the object

        Parameters
        ----------
        root_speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`, optional
            Speaker derived from the root directory, ignored for TextGrids
        sanitize_function: :class:`~montreal_forced_aligner.dictionary.mixins.SanitizeFunction`, optional
            Function to sanitize words and strip punctuation
        """
        if self.text_type == TextFileType.LAB:
            try:
                text = load_text(self.text_path)
            except UnicodeDecodeError:
                raise TextParseError(self.text_path)
            utterance = Utterance(speaker=root_speaker, file=self, text=text)
            utterance.parse_transcription(sanitize_function)
            self.add_utterance(utterance)
        elif self.text_type == TextFileType.TEXTGRID:
            try:
                tg = textgrid.openTextgrid(self.text_path, includeEmptyIntervals=False)
            except Exception:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                raise TextGridParseError(
                    self.text_path,
                    "\n".join(traceback.format_exception(exc_type, exc_value, exc_traceback)),
                )

            num_tiers = len(tg.tierNameList)
            if num_tiers == 0:
                raise TextGridParseError(self.text_path, "Number of tiers parsed was zero")
            if self.num_channels > 2:
                raise (Exception("More than two channels"))
            for tier_name in tg.tierNameList:
                ti = tg.tierDict[tier_name]
                if tier_name.lower() == "notes":
                    continue
                if not isinstance(ti, textgrid.IntervalTier):
                    continue
                if not root_speaker:
                    speaker_name = tier_name.strip()
                    speaker = Speaker(speaker_name)
                    self.add_speaker(speaker)
                else:
                    speaker = root_speaker
                for begin, end, text in ti.entryList:
                    text = text.lower().strip()
                    if not text:
                        continue
                    begin, end = round(begin, 4), round(end, 4)
                    end = min(end, self.duration)
                    utt = Utterance(speaker=speaker, file=self, begin=begin, end=end, text=text)
                    utt.parse_transcription(sanitize_function)
                    if not utt.text:
                        continue
                    self.add_utterance(utt)
        else:
            utterance = Utterance(speaker=root_speaker, file=self)
            self.add_utterance(utterance)

    def add_speaker(self, speaker: Speaker) -> None:
        """
        Add a speaker to a file

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`
            Speaker to add
        """
        if speaker not in self.speaker_ordering:
            self.speaker_ordering.append(speaker)

    def add_utterance(self, utterance: Utterance) -> None:
        """
        Add an utterance to a file

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to add
        """
        self.utterances.add_utterance(utterance)
        self.add_speaker(utterance.speaker)

    def delete_utterance(self, utterance: Utterance) -> None:
        """
        Delete an utterance from the file

        Parameters
        ----------
        utterance: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to remove
        """
        identifier = utterance.name
        del self.utterances[identifier]

    def load_info(self) -> None:
        """
        Load sound file info if it hasn't been already
        """
        if self.wav_path is not None:
            self.wav_info = get_wav_info(self.wav_path)

    @property
    def duration(self) -> float:
        """Get the duration of the sound file"""
        if self.wav_path is None:
            return 0
        if not self.wav_info:
            self.load_info()
        return self.wav_info.duration

    @property
    def num_channels(self) -> int:
        """Get the number of channels of the sound file"""
        if self.wav_path is None:
            return 0
        if not self.wav_info:
            self.load_info()
        return self.wav_info.num_channels

    @property
    def num_utterances(self) -> int:
        """Get the number of utterances for the sound file"""
        return len(self.utterances)

    @property
    def num_speakers(self) -> int:
        """Get the number of speakers in the sound file"""
        return len(self.speaker_ordering)

    @property
    def sample_rate(self) -> int:
        """Get the sample rate of the sound file"""
        if self.wav_path is None:
            return 0
        if not self.wav_info:
            self.load_info()
        return self.wav_info.sample_rate

    @property
    def format(self) -> str:
        """Get the sound file format"""
        if not self.wav_info:
            self.load_info()
        return self.wav_info.format

    @property
    def sox_string(self) -> str:
        """String used for converting sound file via SoX within Kaldi"""
        if not self.wav_info:
            self.load_info()
        return self.wav_info.sox_string

    def load_wav_data(self) -> None:
        """
        Load the sound file into memory as a numpy array
        """
        self.waveform, _ = librosa.load(self.wav_path, sr=None, mono=False)

    def normalized_waveform(
        self, begin: float = 0, end: Optional[float] = None
    ) -> Tuple[np.array, np.array]:
        if self.waveform is None:
            self.load_wav_data()
        if end is None or end > self.duration:
            end = self.duration

        begin_sample = int(begin * self.sample_rate)
        end_sample = int(end * self.sample_rate)
        if len(self.waveform.shape) > 1 and self.waveform.shape[0] == 2:
            y = self.waveform[:, begin_sample:end_sample] / np.max(
                np.abs(self.waveform[:, begin_sample:end_sample]), axis=0
            )
            y[np.isnan(y)] = 0
        else:
            y = self.waveform[begin_sample:end_sample] / np.max(
                np.abs(self.waveform[begin_sample:end_sample]), axis=0
            )
        x = np.arange(start=begin_sample, stop=end_sample) / self.sample_rate
        return x, y

    def for_wav_scp(self) -> str:
        """
        Generate the string to use in feature generation

        Returns
        -------
        str
            SoX string if necessary, the sound file path otherwise
        """
        if self.sox_string:
            return self.sox_string
        return self.wav_path


class Utterance(MfaCorpusClass):
    """
    Class for information about specific utterances

    Parameters
    ----------
    speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`
        Speaker of the utterance
    file: :class:`~montreal_forced_aligner.corpus.classes.File`
        File that the utterance belongs to
    begin: float, optional
        Start time of the utterance,
        if None, then the utterance is assumed to start at 0
    end: float, optional
        End time of the utterance,
        if None, then the utterance is assumed to end at the end of the File
    channel: int, optional
        Channel in the file, if None, then assumed to be the first/only channel
    text: str, optional
        Text transcription of the utterance

    Attributes
    ----------
    file_name: str
        Saved File.name property for reconstructing objects following serialization
    speaker_name: str
        Saved Speaker.name property for reconstructing objects following serialization
    transcription_text: str, optional
        Output of transcription is saved here
    ignored: bool
        The ignored flag is set if feature generation does not work for this utterance, or it is too short to
        be processed by Kaldi
    features: str, optional
        Feature string reference to the computed features archive
    phone_labels: list[:class:`~montreal_forced_aligner.data.CtmInterval`], optional
        Saved aligned phone labels
    word_labels: list[:class:`~montreal_forced_aligner.data.CtmInterval`], optional
        Saved aligned word labels
    oovs: list[str]
        Words not found in the dictionary for this utterance
    """

    def __init__(
        self,
        speaker: Speaker,
        file: File,
        begin: Optional[float] = None,
        end: Optional[float] = None,
        channel: Optional[int] = 0,
        text: Optional[str] = None,
    ):
        self.speaker = speaker
        self.file = file
        self.file_name = file.name
        self.speaker_name: str = speaker.name
        if begin is None:
            begin = 0
        if end is None:
            end = self.file.duration
        self.begin = begin
        self.end = end
        self.channel = channel
        self.text = text
        self.transcription_text = None
        self.ignored = False
        self.features = None
        self.phone_labels: Optional[List[CtmInterval]] = None
        self.word_labels: Optional[List[CtmInterval]] = None
        self.reference_phone_labels: Optional[List[CtmInterval]] = []
        self.oovs = set()
        self.normalized_text = []
        self.text_int = []
        self.alignment_log_likelihood = None
        self.word_error_rate = None
        self.character_error_rate = None
        self.phone_error_rate = None
        self.alignment_score = None

    def parse_transcription(self, sanitize_function=Optional[MultispeakerSanitizationFunction]):
        """
        Parse an orthographic transcription given punctuation and clitic markers

        Parameters
        ----------
        sanitize_function: :class:`~montreal_forced_aligner.dictionary.multispeaker.MultispeakerSanitizationFunction`, optional
            Function to sanitize words and strip punctuation

        """
        self.normalized_text = []
        if sanitize_function is not None:
            try:
                sanitize, split = sanitize_function.get_functions_for_speaker(self.speaker_name)
            except AttributeError:
                sanitize = sanitize_function
                split = None
            words = [
                sanitize(w)
                for w in self.text.split()
                if w and w not in sanitize.clitic_markers + sanitize.compound_markers
            ]
            self.text = " ".join(words)
            if split is not None:
                for w in words:
                    for new_w in split(w):
                        if not new_w:
                            continue
                        if split.word_set is not None and new_w not in split.word_set:
                            self.oovs.add(new_w)
                        self.normalized_text.append(new_w)

    @property
    def multiprocessing_data(self):
        return UtteranceData(
            self.speaker_name,
            self.file_name,
            self.begin,
            self.end,
            self.channel,
            self.text,
            self.normalized_text,
            self.oovs,
        )

    @classmethod
    def load_from_mp_data(cls, data: UtteranceData, file: File) -> Utterance:
        utterance = Utterance(
            speaker=Speaker(data.speaker_name),
            file=file,
            begin=data.begin,
            end=data.end,
            channel=data.channel,
            text=data.text,
        )
        if data.normalized_text:
            utterance.normalized_text = data.normalized_text
            utterance.oovs = data.oovs
        return utterance

    def __str__(self) -> str:
        """String representation"""
        return self.name

    def __repr__(self) -> str:
        """Object representation"""
        return f'<Utterance "{self.name}">'

    def __eq__(self, other: Union[Utterance, str]) -> bool:
        """Check if an Utterance is equal to another Utterance"""
        if isinstance(other, Utterance):
            return other.name == self.name
        if isinstance(other, str):
            return self.name == other
        raise TypeError("Utterances can only be compared to other utterances and strings.")

    def __lt__(self, other: Union[Utterance, str]) -> bool:
        """Check if an Utterance is less than another Utterance"""
        if isinstance(other, Utterance):
            return other.name < self.name
        if isinstance(other, str):
            return self.name < other
        raise TypeError("Utterances can only be compared to other utterances and strings.")

    def __gt__(self, other: Union[Utterance, str]) -> bool:
        """Check if an Utterance is greater than another Utterance"""
        if isinstance(other, Utterance):
            return other.name > self.name
        if isinstance(other, str):
            return self.name > other
        raise TypeError("Utterances can only be compared to other utterances and strings.")

    def __hash__(self) -> hash:
        """Compute the hash of this function"""
        return hash(self.name)

    @property
    def duration(self) -> float:
        """Duration of the utterance"""
        if self.begin is not None and self.end is not None:
            return self.end - self.begin
        return self.file.duration

    @property
    def meta(self) -> Dict[str, Any]:
        """Metadata dictionary for the utterance"""
        return {
            "speaker": self.speaker.name,
            "file": self.file.name,
            "begin": self.begin,
            "end": self.end,
            "channel": self.channel,
            "text": self.text,
            "ignored": self.ignored,
            "features": self.features,
            "normalized_text": self.normalized_text,
            "oovs": self.oovs,
            "transcription_text": self.transcription_text,
            "reference_phone_labels": self.reference_phone_labels,
            "phone_labels": self.phone_labels,
            "word_labels": self.word_labels,
            "word_error_rate": self.word_error_rate,
            "character_error_rate": self.character_error_rate,
            "phone_error_rate": self.phone_error_rate,
            "alignment_score": self.alignment_score,
            "alignment_log_likelihood": self.alignment_log_likelihood,
        }

    def set_speaker(self, speaker: Speaker) -> None:
        """
        Set the speaker of the utterance and updates other objects

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`
            New speaker
        """
        self.speaker = speaker
        self.speaker.add_utterance(self)
        self.file.add_utterance(self)

    @property
    def is_segment(self) -> bool:
        """Check if this utterance is a segment of a longer file"""
        return (
            self.begin is not None
            and self.end is not None
            and self.begin != 0
            and self.end != self.file.duration
        )

    def add_word_intervals(self, intervals: Union[CtmInterval, List[CtmInterval]]) -> None:
        """
        Add aligned word intervals for the utterance

        Parameters
        ----------
        intervals: Union[CtmInterval, list[CtmInterval]]
            Intervals to add
        """
        if not isinstance(intervals, list):
            intervals = [intervals]
        if self.word_labels is None:
            self.word_labels = []
        if self.is_segment:
            for interval in intervals:
                interval.shift_times(self.begin)
        self.word_labels = intervals

    def add_phone_intervals(self, intervals: Union[CtmInterval, List[CtmInterval]]) -> None:
        """
        Add aligned phone intervals for the utterance

        Parameters
        ----------
        intervals: Union[CtmInterval, list[CtmInterval]]
            Intervals to add
        """
        if not isinstance(intervals, list):
            intervals = [intervals]
        if self.phone_labels is None:
            self.phone_labels = []
        if self.is_segment:
            for interval in intervals:
                interval.shift_times(self.begin)
        self.phone_labels = intervals

    def text_for_scp(self) -> List[str]:
        """
        Generate the text for exporting to Kaldi's text scp

        Returns
        -------
        list[str]
            List of words
        """
        return self.text.split()

    def text_int_for_scp(self) -> Optional[List[int]]:
        """
        Generate the text for exporting to Kaldi's text int scp

        Returns
        -------
        list[int]
            List of word IDs, or None if the utterance's speaker doesn't have an associated dictionary
        """
        if self.speaker.dictionary is None:
            return
        if self.text_int:
            return self.text_int
        if self.normalized_text:
            normalized = True
            text = self.normalized_text
        else:
            normalized = False
            text = self.text_for_scp()
        self.text_int = []
        for i, t in enumerate(text):
            lookup = self.speaker.dictionary.to_int(t, normalized)
            if self.speaker.dictionary.oov_int in lookup:
                self.oovs.add(text[i])
            self.text_int.extend(lookup)
        return self.text_int

    def segment_for_scp(self) -> List[Any]:
        """
        Generate data for Kaldi's segments scp file

        Returns
        -------
        list[Any]
            Segment data
        """
        return [self.file_name.replace(" ", "_MFASPACE_"), self.begin, self.end, self.channel]

    @property
    def name(self) -> str:
        """The name of the utterance"""
        base = f"{self.file_name}"
        base = base.replace(" ", "-space-").replace(".", "-").replace("_", "-")
        if not base.startswith(f"{self.speaker_name}-"):
            base = f"{self.speaker_name}-" + base
        if self.is_segment:
            base = f"{base}-{self.begin}-{self.end}"
        return base.replace(" ", "-space-").replace(".", "-").replace("_", "-")


T = TypeVar("T", Speaker, File, Utterance)


class Collection:
    """
    Utility class for storing collections of corpus objects, allowing iteration, sorting, and
    look up via names.
    """

    CLASS_TYPE = ClassVar[MfaCorpusClass]

    def __init__(self):
        self._data: Dict[str, T] = {}

    def __getitem__(self, key: str) -> T:
        """Get an item by identifier"""
        return self._data[key]

    def __delitem__(self, key: str) -> None:
        """Delete an item by identifier"""
        del self._data[key]

    def __setitem__(self, key: str, item: T) -> None:
        """Set an item by identifier"""
        self._data[key] = item

    def __len__(self) -> int:
        """Number of items in the collection"""
        return len(self._data)

    def __bool__(self) -> bool:
        """Check for whether the collection contains any items"""
        return bool(self._data)

    def __contains__(self, item: Union[str, T]) -> bool:
        """Check for whether the collection contains a specific item"""
        if not isinstance(item, str):
            item = item.name
        return item in self._data

    def subset(self, subset_identifiers: Set[str]) -> Generator[T]:
        for item in self:
            if subset_identifiers and item.name not in subset_identifiers:
                continue
            yield item

    def __iter__(self) -> Generator[T]:
        """Iterator over the collection"""
        for v in self._data.values():
            yield v

    def update(self, other: Union[Collection, Set[T], List[T]]) -> None:
        """Update collection from another collection"""
        if isinstance(other, Collection):
            self._data.update(other._data)
        else:
            for item in other:
                self._data[item.name] = item

    def __str__(self) -> str:
        """String representation"""
        return str(self._data)

    def __repr__(self) -> str:
        """Object representation"""
        return f"<Collection of {self._data}>"


class SpeakerCollection(Collection):
    """
    Utility class for storing collections of speakers
    """

    CLASS_TYPE = Speaker

    def add_speaker(self, speaker: Speaker) -> None:
        """
        Add speaker to the collection

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`
            Speaker to be added
        """
        self[speaker.name] = speaker

    def __repr__(self) -> str:
        """Object representation"""
        return f"<SpeakerCollection of {self._data}>"


class FileCollection(Collection):
    """
    Utility class for storing collections of speakers
    """

    CLASS_TYPE = File

    def __init__(self):
        super(FileCollection, self).__init__()
        self.lab_count = 0
        self.textgrid_count = 0
        self.sound_file_count = 0

    def add_file(self, file: File) -> None:
        """
        Add file to the collection

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.File`
            File to be added
        """
        self[file.name] = file
        if file.text_type == TextFileType.TEXTGRID:
            self.textgrid_count += 1
        elif file.text_type == TextFileType.LAB:
            self.lab_count += 1
        if file.sound_type != SoundFileType.NONE:
            self.sound_file_count += 1

    def __repr__(self) -> str:
        """Object representation"""
        return f"<FileCollection of {self._data}>"


class UtteranceCollection(Collection):
    """
    Utility class for storing collections of speakers
    """

    CLASS_TYPE = Utterance

    def add_utterance(self, utterance: Utterance) -> None:
        """
        Add utterance to the collection

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.Utterance`
            Utterance to be added
        """
        self[utterance.name] = utterance

    def __iter__(self) -> Generator[Utterance]:
        """Iterator over the collection"""
        for v in self._data.values():
            if v.ignored:
                continue
            yield v

    def __repr__(self) -> str:
        """Object representation"""
        return f"<UtteranceCollection of {self._data}>"



"""
Data classes
============

"""
from __future__ import annotations

import dataclasses
import enum
import itertools
import re
import typing

from praatio.utilities.constants import Interval

from .exceptions import CtmError

__all__ = [
    "CtmInterval",
    "UtteranceData",
    "FileData",
    "TextFileType",
    "SoundFileType",
    "PhoneSetType",
]


class TextFileType(enum.Enum):
    """Enum for types of text files"""

    NONE = 0
    TEXTGRID = 1
    LAB = 2


class SoundFileType(enum.Enum):
    """Enum for types of sound files"""

    NONE = 0
    WAV = 1
    SOX = 2


def voiceless_variants(base_phone):
    return {base_phone + d for d in ["", "ʱ", "ʼ", "ʰ", "ʲ", "ʷ", "ˠ", "ˀ", "̚", "͈"]}


def voiced_variants(base_phone):
    return {base_phone + d for d in ["", "ʱ", "ʲ", "ʷ", "ⁿ", "ˠ", "̚"]} | {
        d + base_phone for d in ["ⁿ"]
    }


class PhoneSetType(enum.Enum):
    """Enum for types of phone sets"""

    UNKNOWN = "UNKNOWN"
    AUTO = "AUTO"
    IPA = "IPA"
    ARPA = "ARPA"
    PINYIN = "PINYIN"

    def __str__(self):
        """Name of phone set"""
        return self.name

    @property
    def has_base_phone_regex(self):
        return self is PhoneSetType.IPA or self is PhoneSetType.ARPA or self is PhoneSetType.PINYIN

    @property
    def regex_detect(self) -> typing.Optional[re.Pattern]:
        if self is PhoneSetType.ARPA:
            return re.compile(r"[A-Z]{2}[012]")
        elif self is PhoneSetType.PINYIN:
            return re.compile(r"[a-z]{1,3}[12345]")
        elif self is PhoneSetType.IPA:
            return re.compile(
                r"[əɚʊɡɤʁɹɔɛʉɒβɲɟʝŋʃɕʰʲɾ̃̚ː˩˨˧˦˥̪̝̟̥̂̀̄ˑ̊ᵝ̠̹̞̩̯̬̺ˀˤ̻̙̘̰̤̜̹̑̽᷈᷄᷅̌̋̏‿̆͜͡ˌˈ̣]"
            )
        return None

    def get_base_phone(self, phone: str):
        if self.has_base_phone_regex:
            return self.base_phone_regex.sub("", phone)
        return phone

    @property
    def base_phone_regex(self) -> typing.Optional[re.Pattern]:
        if self is PhoneSetType.ARPA:
            return re.compile(r"[012]")
        elif self is PhoneSetType.PINYIN:
            return re.compile(r"[12345]")
        elif self is PhoneSetType.IPA:
            return re.compile(r"([ː˩˨˧˦˥̪̟̥̂̀̄ˑ̊ᵝ̠̹̞̩̯̬̺ˀˤ̻̙̘̤̜̹̑̽᷈᷄᷅̌̋̏‿̆͜͡ˌˈ̣]+)")
        return None

    @property
    def extra_short_phones(self) -> typing.Set[str]:
        if self is PhoneSetType.ARPA:
            return {"AH0", "IH0", "ER0", "UH0"}
        elif self is PhoneSetType.IPA:
            return {"ʔ", "ə", "ɚ", "ɾ", "p̚", "t̚", "k̚"}
        return set()

    @property
    def affricate_phones(self) -> typing.Set[str]:
        if self is PhoneSetType.ARPA:
            return {"CH", "JH"}
        if self is PhoneSetType.PINYIN:
            return {"z", "zh", "j", "c", "ch", "q"}
        if self is PhoneSetType.IPA:
            affricates = set()
            for p in {
                "pf",
                "ts",
                "tʃ",
                "tɕ",
                "tʂ",
                "ʈʂ",
                "cç",
                "kx",
                "tç",
            }:
                affricates |= voiceless_variants(p)
            for p in {
                "dz",
                "dʒ",
                "dʑ",
                "dʐ",
                "ɖʐ",
                "ɟʝ",
                "ɡɣ",
                "dʝ",
            }:
                affricates |= voiced_variants(p)
            return affricates
        return set()

    @property
    def stop_phones(self) -> typing.Set[str]:
        if self is PhoneSetType.ARPA:
            return {"B", "D", "G"}
        if self is PhoneSetType.PINYIN:
            return {"b", "d", "g"}
        if self is PhoneSetType.IPA:
            stops = set()
            for p in {"p", "t", "ʈ", "c", "k", "q"}:
                stops |= voiceless_variants(p)
            for p in {"b", "d", "ɖ", "ɟ", "ɡ", "ɢ"}:
                stops |= voiced_variants(p)
            return stops
        return set()

    @property
    def diphthong_phones(self) -> typing.Set[str]:
        if self is PhoneSetType.ARPA:
            return {
                "AY0",
                "AY1",
                "AY2",
                "AW0",
                "AW1",
                "AW2",
                "OY0",
                "OY1",
                "OY2",
                "EY0",
                "EY1",
                "EY2",
                "OW0",
                "OW1",
                "OW2",
            }
        if self is PhoneSetType.IPA or self is PhoneSetType.PINYIN:
            return {x + y for x, y in itertools.product(self.vowels, self.vowels)}

        return set()

    @property
    def vowels(self):
        if self is PhoneSetType.PINYIN:
            return {"i", "u", "y", "e", "w", "a", "o", "e", "ü"}
        if self is PhoneSetType.IPA:
            base_vowels = {
                "i",
                "u",
                "e",
                "ə",
                "a",
                "o",
                "y",
                "ɔ",
                "j",
                "w",
                "ɪ",
                "ʊ",
                "w",
                "ʏ",
                "ɯ",
                "ɤ",
                "ɑ",
                "æ",
                "ɐ",
                "ɚ",
                "ɵ",
                "ɘ",
                "ɛ",
                "ɜ",
                "ɝ",
                "ɞ",
                "ɑ̃",
                "ɨ",
                "ɪ̈",
                "œ",
                "ɒ",
                "ɶ",
                "ø",
                "ʉ",
                "ʌ",
            }
            base_vowels |= {x + "̃" for x in base_vowels}  # Add nasals
            return {
                "i",
                "u",
                "e",
                "ə",
                "a",
                "o",
                "y",
                "ɔ",
                "j",
                "w",
                "ɪ",
                "ʊ",
                "w",
                "ʏ",
                "ɯ",
                "ɤ",
                "ɑ",
                "æ",
                "ɐ",
                "ɚ",
                "ɵ",
                "ɘ",
                "ɛ",
                "ɜ",
                "ɝ",
                "ɞ",
                "ɨ",
                "ɪ̈",
                "œ",
                "ɒ",
                "ɶ",
                "ø",
                "ʉ",
                "ʌ",
            }
        return set()

    @property
    def triphthong_phones(self) -> typing.Set[str]:
        if self is PhoneSetType.IPA or self is PhoneSetType.PINYIN:
            return {
                x + y + z for x, y, z in itertools.product(self.vowels, self.vowels, self.vowels)
            }

        return set()

    @property
    def extra_questions(self) -> typing.Dict[str, typing.Set[str]]:
        extra_questions = {}
        if self is PhoneSetType.ARPA:
            extra_questions["bilabial_variation"] = {"P", "B"}
            extra_questions["dental_lenition"] = {"D", "DH"}
            extra_questions["flapping"] = {"T", "D"}
            extra_questions["nasal_variation"] = {"M", "N", "NG"}
            extra_questions["voiceless_sibilant_variation"] = {"CH", "SH", "S"}
            extra_questions["voiceless_sibilant_variation"] = {"JH", "ZH", "Z"}
            extra_questions["voiceless_fricative_variation"] = {"F", "TH", "HH", "K"}
            extra_questions["voiced_fricative_variation"] = {"V", "DH", "HH", "G"}
            extra_questions["dorsal_variation"] = {"HH", "K", "G"}
            extra_questions["rhotic_variation"] = {"ER0", "ER1", "ER2", "R"}

            extra_questions["low_back_variation"] = {
                "AO0",
                "AO1",
                "AO2",
                "AA0",
                "AA1",
                "AA2",
            }
            extra_questions["central_variation"] = {
                "ER0",
                "ER1",
                "ER2",
                "AH0",
                "AH1",
                "AH2",
                "UH0",
                "UH1",
                "UH2",
                "IH0",
                "IH1",
                "IH2",
            }
            extra_questions["high_back_variation"] = {
                "UW1",
                "UW2",
                "UW0",
                "UH1",
                "UH2",
                "UH0",
            }
            extra_questions["high_front_variation"] = {
                "IY1",
                "IY2",
                "IY0",
                "IH0",
                "IH1",
                "IH2",
            }
            extra_questions["mid_front_variation"] = {
                "EY1",
                "EY2",
                "EY0",
                "EH0",
                "EH1",
                "EH2",
            }

            # extra stress questions
            vowels = [
                "AA",
                "AE",
                "AH",
                "AO",
                "AW",
                "AY",
                "EH",
                "ER",
                "EY",
                "IH",
                "IY",
                "OW",
                "OY",
                "UH",
                "UW",
            ]
            for i in range(3):
                extra_questions[f"stress_{i}"] = {f"{x}{i}" for x in vowels}
        elif self is PhoneSetType.PINYIN:
            for i in range(1, 6):
                extra_questions[f"tone_{i}"] = {f"{x}{i}" for x in self.vowels}
                extra_questions[f"tone_{i}"] |= {f"{x}{i}" for x in self.diphthong_phones}
                extra_questions[f"tone_{i}"] |= {f"{x}{i}" for x in self.triphthong_phones}
            extra_questions["bilabial_variation"] = {"p", "b"}
            extra_questions["nasal_variation"] = {"m", "n", "ng"}
            extra_questions["voiceless_sibilant_variation"] = {
                "z",
                "zh",
                "j",
                "c",
                "ch",
                "q",
                "s",
                "sh",
                "x",
            }
            extra_questions["dorsal_variation"] = {"h", "k", "g"}
            extra_questions["alveolar_stop_variation"] = {"t", "d"}
            extra_questions["approximant_variation"] = {"l", "r", "y", "w"}
            extra_questions["rhotic_variation"] = {"r", "sh", "e"}

        elif self is PhoneSetType.IPA:

            extra_questions["dental_lenition"] = voiced_variants("ð") | voiced_variants("d")
            extra_questions["flapping"] = {"d", "t", "ɾ"}
            extra_questions["glottalization"] = {"t", "ʔ", "t̚"}
            extra_questions["labial_lenition"] = voiced_variants("β") | voiced_variants("b")
            extra_questions["velar_lenition"] = voiced_variants("ɣ") | voiced_variants("ɡ")

            nasal_variation = (
                voiced_variants("m")
                | voiced_variants("n")
                | voiced_variants("ɲ")
                | voiced_variants("ŋ")
            )
            nasal_variation |= voiced_variants("ɴ") | voiced_variants("ɳ") | voiced_variants("ɱ")
            nasal_variation |= voiced_variants("ɰ̃") | voiced_variants("ɾ")
            extra_questions["nasal_variation"] = nasal_variation

            approximant_variation = (
                voiced_variants("w")
                | voiced_variants("ʍ")
                | voiced_variants("ɰ")
                | voiced_variants("ɥ")
            )
            approximant_variation |= (
                voiced_variants("l")
                | voiced_variants("ɭ")
                | voiced_variants("ɹ")
                | voiced_variants("ɫ")
            )
            approximant_variation |= (
                voiced_variants("ɻ") | voiced_variants("ʋ") | voiced_variants("j")
            )
            extra_questions["approximant_variation"] = approximant_variation

            front_glide_variation = {"j", "i", "ɪ", "ɥ", "ʏ", "y"}
            front_glide_variation |= {x + "̃" for x in front_glide_variation}
            extra_questions["front_glide_variation"] = front_glide_variation

            back_glide_variation = {"w", "u", "ʊ", "ɰ", "ɯ", "ʍ"}
            back_glide_variation |= {x + "̃" for x in back_glide_variation}
            extra_questions["back_glide_variation"] = back_glide_variation

            trill_variation = (
                voiced_variants("r")
                | voiced_variants("ɾ")
                | voiced_variants("ɽ")
                | voiced_variants("ɽr")
            )
            trill_variation |= voiced_variants("ʁ") | voiced_variants("ʀ") | voiced_variants("ɢ̆")
            trill_variation |= voiced_variants("ɺ") | voiced_variants("ɭ")
            extra_questions["trill_variation"] = trill_variation

            extra_questions["syllabic_rhotic_variation"] = {"ɹ", "ɝ", "ɚ", "ə", "ʁ", "ɐ"}
            uvular_variation = (
                voiced_variants("ʁ") | voiceless_variants("x") | voiceless_variants("χ")
            )
            uvular_variation |= (
                voiced_variants("ɣ")
                | voiceless_variants("h")
                | voiced_variants("ɰ")
                | voiced_variants("ʀ")
            )
            extra_questions["uvular_variation"] = uvular_variation

            lateral_variation = voiced_variants("l") | voiced_variants("ɫ") | voiced_variants("ʎ")
            lateral_variation |= (
                voiced_variants("ʟ")
                | voiced_variants("ɭ")
                | voiced_variants("ɮ")
                | voiceless_variants("ɬ")
            )
            extra_questions["lateral_variation"] = lateral_variation

            dorsal_stops = voiceless_variants("k") | voiced_variants("ɡ") | voiced_variants("ɠ")
            dorsal_stops |= voiceless_variants("c") | voiced_variants("ɟ") | voiced_variants("ʄ")
            dorsal_stops |= voiceless_variants("q") | voiced_variants("ɢ") | voiced_variants("ʛ")
            extra_questions["dorsal_stop_variation"] = dorsal_stops

            bilabial_stops = (
                voiceless_variants("p")
                | voiced_variants("b")
                | voiced_variants("ɓ")
                | voiced_variants("ʙ")
                | voiced_variants("ⱱ")
            )
            extra_questions["bilabial_stop_variation"] = bilabial_stops

            alveolar_stops = voiceless_variants("t") | voiceless_variants("ʈ")
            alveolar_stops |= (
                voiced_variants("d")
                | voiced_variants("ɗ")
                | voiced_variants("ɖ")
                | voiced_variants("ᶑ")
            )
            extra_questions["alveolar_stop_variation"] = alveolar_stops

            voiceless_fricatives = (
                voiceless_variants("θ")
                | voiceless_variants("f")
                | voiceless_variants("pf")
                | voiceless_variants("ɸ")
            )
            voiceless_fricatives |= (
                voiceless_variants("ç")
                | voiceless_variants("x")
                | voiceless_variants("kx")
                | voiceless_variants("χ")
                | voiceless_variants("h")
            )
            voiceless_fricatives |= (
                voiceless_variants("cç")
                | voiceless_variants("q")
                | voiceless_variants("qχ")
                | voiceless_variants("ɬ")
            )
            extra_questions["voiceless_fricative_variation"] = voiceless_fricatives
            voiced_fricatives = voiced_variants("v") | voiced_variants("ð")
            voiced_fricatives |= voiced_variants("β") | voiced_variants("ʋ") | voiced_variants("ɣ")
            voiced_fricatives |= (
                voiced_variants("ɟʝ") | voiced_variants("ʝ") | voiced_variants("ɮ")
            )

            extra_questions["voiced_fricative_variation"] = voiced_fricatives
            voiceless_sibilants = (
                voiceless_variants("s")
                | voiceless_variants("ʃ")
                | voiceless_variants("ɕ")
                | voiceless_variants("ʂ")
            )
            voiceless_sibilants |= (
                voiceless_variants("ts") | voiceless_variants("tʃ") | voiceless_variants("tɕ")
            )
            voiceless_sibilants |= voiceless_variants("ʈʂ") | voiceless_variants("tʂ")
            extra_questions["voiceless_sibilant_variation"] = voiceless_sibilants

            voiced_sibilants = (
                voiced_variants("z")
                | voiced_variants("ʒ")
                | voiced_variants("ʐ")
                | voiced_variants("ʑ")
            )
            voiced_sibilants |= (
                voiced_variants("dz")
                | voiced_variants("dʒ")
                | voiced_variants("dʐ")
                | voiced_variants("ɖʐ")
                | voiced_variants("dʑ")
            )
            extra_questions["voiced_sibilant_variation"] = voiced_sibilants

            extra_questions["low_vowel_variation"] = {"a", "ɐ", "ɑ", "ɔ"}
            extra_questions["mid_back_vowel_variation"] = {"oʊ", "ɤ", "o", "ɔ"}
            extra_questions["mid_front_variation"] = {"ɛ", "eɪ", "e", "œ", "ø"}
            extra_questions["high_front_variation"] = {"i", "y", "ɪ", "ʏ", "ɨ", "ʉ"}
            extra_questions["high_back_variation"] = {"ʊ", "u", "ɯ", "ɨ", "ʉ"}
            back_diphthong_variation = {"aʊ", "au", "aw", "ɐw"}
            back_diphthong_variation |= {x + "̃" for x in back_diphthong_variation}
            extra_questions["back_diphthong_variation"] = back_diphthong_variation

            front_diphthong_variation = {"aɪ", "ai", "aj", "ɐj"}
            front_diphthong_variation |= {x + "̃" for x in front_diphthong_variation}
            extra_questions["front_diphthong_variation"] = front_diphthong_variation

            mid_front_diphthong_variation = {
                "eɪ",
                "ei",
                "ej",
                "ɛɪ",
                "ɛi",
                "ɛj",
            }
            mid_front_diphthong_variation |= {x + "̃" for x in mid_front_diphthong_variation}
            extra_questions["mid_front_diphthong_variation"] = mid_front_diphthong_variation

            mid_back_diphthong_variation = {"ow", "ou", "oʊ"}
            mid_back_diphthong_variation |= {x + "̃" for x in mid_back_diphthong_variation}
            extra_questions["mid_front_diphthong_variation"] = mid_back_diphthong_variation

            mid_back_front_diphthong_variation = {"oj", "oi", "oɪ", "ɔʏ", "ɔɪ", "ɔj", "ɔi"}
            mid_back_front_diphthong_variation |= {
                x + "̃" for x in mid_back_front_diphthong_variation
            }
            extra_questions[
                "mid_back_front_diphthong_variation"
            ] = mid_back_front_diphthong_variation

            extra_questions["central_variation"] = {
                "ə",
                "ɤ",
                "ɚ",
                "ʌ",
                "ʊ",
                "ɵ",
                "ɐ",
                "ɞ",
                "ɘ",
                "ɝ",
            }
        return extra_questions


@dataclasses.dataclass
class SoundFileInformation:
    """
    Data class for sound file information with format, duration, number of channels, bit depth, and
        sox_string for use in Kaldi feature extraction if necessary

    Parameters
    ----------
    format: str
        Format of the sound file
    sample_rate: int
        Sample rate
    duration: float
        Duration
    sample_rate: int
        Sample rate
    bit_depth: int
        Bit depth
    sox_string: str
        String to use for loading with sox
    """

    __slots__ = ["format", "sample_rate", "duration", "num_channels", "bit_depth", "sox_string"]
    format: str
    sample_rate: int
    duration: float
    num_channels: int
    bit_depth: int
    sox_string: str

    @property
    def meta(self) -> typing.Dict[str, typing.Any]:
        return dataclasses.asdict(self)


@dataclasses.dataclass
class FileExtensions:
    """
    Data class for information about the current directory

    Parameters
    ----------
    identifiers: list[str]
        List of identifiers
    lab_files: dict[str, str]
        Mapping of identifiers to lab files
    textgrid_files: dict[str, str]
        Mapping of identifiers to TextGrid files
    wav_files: dict[str, str]
        Mapping of identifiers to wav files
    other_audio_files: dict[str, str]
        Mapping of identifiers to other audio files
    """

    __slots__ = ["identifiers", "lab_files", "textgrid_files", "wav_files", "other_audio_files"]

    identifiers: typing.List[str]
    lab_files: typing.Dict[str, str]
    textgrid_files: typing.Dict[str, str]
    wav_files: typing.Dict[str, str]
    other_audio_files: typing.Dict[str, str]


@dataclasses.dataclass
class Pronunciation:
    """
    Data class for information about a pronunciation string

    Parameters
    ----------
    pronunciation: tuple
        Tuple of phones
    probability: float
        Probability of pronunciation
    disambiguation: Optional[int]
        Disambiguation index within a pronunciation dictionary
    left_silence_probability: Optional[float]
        Probability of silence before the pronunciation
    right_silence_probability: Optional[float]
        Probability of silence after the pronunciation
    """

    __slots__ = [
        "pronunciation",
        "probability",
        "disambiguation",
        "left_silence_probability",
        "right_silence_probability",
    ]

    pronunciation: typing.Tuple[str, ...]
    probability: float
    disambiguation: typing.Optional[int]
    left_silence_probability: typing.Optional[float]
    right_silence_probability: typing.Optional[float]

    def __hash__(self):
        return hash(self.pronunciation)

    def __len__(self):
        return len(self.pronunciation)

    def __repr__(self):
        return f"<Pronunciation /{' '.join(self.pronunciation)}/>"

    def __bool__(self) -> bool:
        return bool(self.pronunciation)

    def __str__(self):
        return f"{' '.join(self.pronunciation)}"

    def __eq__(self, other: Pronunciation):
        return self.pronunciation == other.pronunciation

    def __lt__(self, other: Pronunciation):
        return self.pronunciation < other.pronunciation

    def __gt__(self, other: Pronunciation):
        return self.pronunciation > other.pronunciation


@dataclasses.dataclass
class Word:
    """
    Data class for information about a word and its pronunciations

    Parameters
    ----------
    orthography: str
        Orthographic string for the word
    pronunciations: set[:class:`~montreal_forced_aligner.data.Pronunciation`]
        Set of pronunciations for the word
    """

    __slots__ = ["orthography", "pronunciations"]

    orthography: str
    pronunciations: typing.Set[Pronunciation]

    def __repr__(self) -> str:
        """Word object representation"""
        pronunciation_string = ", ".join(map(str, self.pronunciations))
        return f"<Word {self.orthography} with pronunciations {pronunciation_string}>"

    def __hash__(self) -> hash:
        """Word hash"""
        return hash(self.orthography)

    def __len__(self) -> int:
        """Number of pronunciations"""
        return len(self.pronunciations)

    def __iter__(self) -> typing.Generator[Pronunciation]:
        """Iterator over pronunciations"""
        for p in self.pronunciations:
            yield p


@dataclasses.dataclass
class UtteranceData:
    """
    Data class for utterance information

    Parameters
    ----------
    speaker_name: str
        Speaker name
    file_name: str
        File name
    begin: float, optional
        Begin timestamp
    end: float, optional
        End timestamp
    channel: int, optional
        Sound file channel
    text: str, optional
        Utterance text
    normalized_text: list[str]
        Normalized utterance text, with compounds and clitics split up
    oovs: set[str]
        Set of words not found in a look up
    """

    __slots__ = [
        "speaker_name",
        "file_name",
        "begin",
        "end",
        "channel",
        "text",
        "normalized_text",
        "oovs",
    ]
    speaker_name: str
    file_name: str
    begin: typing.Optional[float]
    end: typing.Optional[float]
    channel: typing.Optional[int]
    text: typing.Optional[str]
    normalized_text: typing.List[str]
    oovs: typing.Set[str]


@dataclasses.dataclass
class FileData:
    """
    Data class for file information

    Parameters
    ----------
    name: str
        File name
    wav_path: str, optional
        Path to sound file
    text_path: str, optional
        Path to sound file
    relative_path: str
        Path relative to corpus root directory
    wav_info: dict[str, Any]
        Information dictionary about the sound file
    speaker_ordering: list[str]
        List of speakers in the file
    utterances: list[:class:`~montreal_forced_aligner.data.UtteranceData`]
        Utterance data for the file
    """

    __slots__ = [
        "name",
        "wav_path",
        "text_path",
        "relative_path",
        "wav_info",
        "speaker_ordering",
        "utterances",
    ]
    name: str
    wav_path: typing.Optional[str]
    text_path: typing.Optional[str]
    relative_path: str
    wav_info: SoundFileInformation
    speaker_ordering: typing.List[str]
    utterances: typing.List[UtteranceData]


@dataclasses.dataclass
class CtmInterval:
    """
    Data class for intervals derived from CTM files

    Parameters
    ----------
    begin: float
        Start time of interval
    end: float
        End time of interval
    label: str
        Text of interval
    utterance: str
        Utterance ID that the interval belongs to
    """

    __slots__ = ["begin", "end", "label", "utterance"]

    begin: float
    end: float
    label: str
    utterance: str

    def __lt__(self, other: CtmInterval):
        return self.begin < other.begin

    def __post_init__(self):
        """
        Check on data validity

        Raises
        ------
        :class:`~montreal_forced_aligner.exceptions.CtmError`
            If begin or end are not valid
        """
        if self.end < -1 or self.begin == 1000000:
            raise CtmError(self)

    def __add__(self, o: str):
        return self.label + o

    def shift_times(self, offset: float):
        """
        Shift times of the interval based on some offset (i.e., segments in Kaldi)

        Parameters
        ----------
        offset: float
            Offset to add to the interval's begin and end
        """
        self.begin += offset
        self.end += offset

    def to_tg_interval(self) -> Interval:
        """
        Converts the CTMInterval to `PraatIO's Interval class <http://timmahrt.github.io/praatIO/praatio/utilities/constants.html#Interval>`_

        Returns
        -------
        :class:`praatio.utilities.constants.Interval`
            Derived PraatIO Interval
        """
        if self.end < -1 or self.begin == 1000000:
            raise CtmError(self)
        return Interval(round(self.begin, 4), round(self.end, 4), self.label)



"""
Corpus loading worker
---------------------
"""
from __future__ import annotations

import multiprocessing as mp
import os
import sys
import traceback
from queue import Empty
from typing import TYPE_CHECKING, Any, Collection, Dict, Generator, List, Optional, Set, Union

from montreal_forced_aligner.corpus.classes import (
    File,
    FileCollection,
    SpeakerCollection,
    Utterance,
    UtteranceCollection,
)
from montreal_forced_aligner.dictionary.multispeaker import MultispeakerSanitizationFunction
from montreal_forced_aligner.exceptions import TextGridParseError, TextParseError
from montreal_forced_aligner.helper import output_mapping
from montreal_forced_aligner.utils import Stopped

if TYPE_CHECKING:
    from montreal_forced_aligner.abc import OneToManyMappingType, OneToOneMappingType
    from montreal_forced_aligner.corpus.helper import SoundFileInfoDict

    FileInfoDict = Dict[
        str, Union[str, SoundFileInfoDict, OneToOneMappingType, OneToManyMappingType]
    ]
    from montreal_forced_aligner.corpus.classes import Speaker
    from montreal_forced_aligner.dictionary import PronunciationDictionaryMixin


__all__ = ["CorpusProcessWorker", "Job"]


class CorpusProcessWorker(mp.Process):
    """
    Multiprocessing corpus loading worker

    Attributes
    ----------
    job_q: :class:`~multiprocessing.Queue`
        Job queue for files to process
    return_dict: dict
        Dictionary to catch errors
    return_q: :class:`~multiprocessing.Queue`
        Return queue for processed Files
    stopped: :class:`~montreal_forced_aligner.utils.Stopped`
        Stop check for whether corpus loading should exit
    finished_adding: :class:`~montreal_forced_aligner.utils.Stopped`
        Signal that the main thread has stopped adding new files to be processed
    """

    def __init__(
        self,
        name: int,
        job_q: mp.Queue,
        return_dict: dict,
        return_q: mp.Queue,
        stopped: Stopped,
        finished_adding: Stopped,
        speaker_characters: Union[int, str],
        sanitize_function: Optional[MultispeakerSanitizationFunction],
    ):
        mp.Process.__init__(self)
        self.name = str(name)
        self.job_q = job_q
        self.return_dict = return_dict
        self.return_q = return_q
        self.stopped = stopped
        self.finished_adding = finished_adding
        self.finished_processing = Stopped()
        self.sanitize_function = sanitize_function
        self.speaker_characters = speaker_characters

    def run(self) -> None:
        """
        Run the corpus loading job
        """

        while True:
            try:
                file_name, wav_path, text_path, relative_path = self.job_q.get(timeout=1)
            except Empty:
                if self.finished_adding.stop_check():
                    break
                continue
            self.job_q.task_done()
            if self.stopped.stop_check():
                continue
            try:
                file = File.parse_file(
                    file_name,
                    wav_path,
                    text_path,
                    relative_path,
                    self.speaker_characters,
                    self.sanitize_function,
                )

                self.return_q.put(file.multiprocessing_data)
            except TextParseError as e:
                self.return_dict["decode_error_files"].append(e)
            except TextGridParseError as e:
                self.return_dict["textgrid_read_errors"][e.file_name] = e
            except Exception:
                self.stopped.stop()
                self.return_dict["error"] = file_name, Exception(
                    traceback.format_exception(*sys.exc_info())
                )
        self.finished_processing.stop()
        return


class Job:
    """
    Class representing information about corpus jobs that will be run in parallel.
    Jobs have a set of speakers that they will process, along with all files and utterances associated with that speaker.
    As such, Jobs also have a set of dictionaries that the speakers use, and argument outputs are largely dependent on
    the pronunciation dictionaries in use.

    Parameters
    ----------
    name: int
        Job number is the job's identifier

    Attributes
    ----------
    speakers: list[:class:`~montreal_forced_aligner.corpus.classes.Speaker`]
        List of speakers associated with this job
    dictionaries: set[:class:`~montreal_forced_aligner.dictionary.PronunciationDictionary`]
        Set of dictionaries that the job's speakers use
    subset_utts: set[:class:`~montreal_forced_aligner.corpus.classes.Utterance`]
        When trainers are just using a subset of the corpus, the subset of utterances on each job will be set and used to
        filter the job's utterances
    subset_speakers: set[:class:`~montreal_forced_aligner.corpus.classes.Speaker`]
        When subset_utts is set, this property will be calculated as the subset of speakers that the utterances correspond to
    subset_dictionaries: set[:class:`~montreal_forced_aligner.dictionary.PronunciationDictionary`]
        Subset of dictionaries that the subset of speakers use

    """

    name: int
    speakers: SpeakerCollection
    dictionaries: Set[PronunciationDictionaryMixin]

    def __init__(self, name: int):
        self.name = name
        self.speakers = SpeakerCollection()
        self.files = FileCollection()
        self.dictionaries = set()

        self.subset_utts = set()
        self.subset_files = set()
        self.subset_speakers = set()
        self.subset_dictionaries = set()

    def add_speaker(self, speaker: Speaker) -> None:
        """
        Add a speaker to a job

        Parameters
        ----------
        speaker: :class:`~montreal_forced_aligner.corpus.classes.Speaker`
            Speaker to add
        """
        self.speakers.add_speaker(speaker)
        for u in speaker.utterances:
            self.files.add_file(u.file)
        self.dictionaries.add(speaker.dictionary)

    def set_subset(self, subset_utts: Optional[UtteranceCollection]) -> None:
        """
        Set the current subset for the trainer

        Parameters
        ----------
        subset_utts: Collection[:class:`~montreal_forced_aligner.corpus.classes.Utterance`], optional
            Subset of utterances for this job to use
        """
        self.subset_utts = set()
        self.subset_files = set()
        self.subset_speakers = set()
        self.subset_dictionaries = set()
        if subset_utts:
            for u in subset_utts:
                if u.speaker not in self.speakers:
                    continue
                self.subset_utts.add(u.name)
                self.subset_speakers.add(u.speaker_name)
                self.subset_files.add(u.file_name)
                dict_name = self.speakers[u.speaker_name].dictionary_name
                if dict_name is not None:
                    self.subset_dictionaries.add(dict_name)

    def text_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's text scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Text for each utterance, per dictionary name
        """
        data = {x: {} for x in self.current_dictionary_names}
        for utt in self.current_utterances:
            dict_name = utt.speaker.dictionary_name
            if not utt.text:
                continue
            data[dict_name][utt.name] = utt.text
        return data

    def text_int_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's text int scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Text converted to integer IDs for each utterance, per dictionary name
        """
        data = {x: {} for x in self.current_dictionary_names}
        for utt in self.current_utterances:
            dict_name = utt.speaker.dictionary_name
            if dict_name is None:
                continue
            if not utt.text:
                continue
            data[dict_name][utt.name] = " ".join(map(str, utt.text_int_for_scp()))
            utt.speaker.dictionary.oovs_found.update(utt.oovs)
        return data

    def wav_scp_data(self) -> Dict[str, str]:
        """
        Generate the job's data for Kaldi's wav scp files

        Returns
        -------
        dict[str, str]
            Wav scp strings for each file, per dictionary name
        """
        data = {}
        for file in self.files:
            if any(u.is_segment for u in file.utterances):  # Segmented file
                data[file.name] = file.for_wav_scp()
            else:
                for utt in file.utterances:
                    data[utt.name] = file.for_wav_scp()
        return data

    def utt2spk_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's utt2spk scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Utterance to speaker mapping, per dictionary name
        """
        data = {x: {} for x in self.current_dictionary_names}
        for utt in self.current_utterances:
            dict_name = utt.speaker.dictionary_name
            data[dict_name][utt.name] = utt.speaker_name
        return data

    def feat_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's feature scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Utterance to feature archive ID mapping, per dictionary name
        """
        data = {x: {} for x in self.current_dictionary_names}
        for utt in self.current_utterances:
            if utt.features is None:
                continue
            dict_name = utt.speaker.dictionary_name
            data[dict_name][utt.name] = utt.features
        return data

    def spk2utt_scp_data(self) -> Dict[str, Dict[str, List[str]]]:
        """
        Generate the job's data for Kaldi's spk2utt scp files

        Returns
        -------
        dict[str, dict[str, list[str]]]
            Speaker to utterance mapping, per dictionary name
        """
        data = {x: {} for x in self.current_dictionary_names}
        for utt in self.current_utterances:
            dict_name = utt.speaker.dictionary_name
            if utt.speaker_name not in data[dict_name]:
                data[dict_name][utt.speaker_name] = []
            data[dict_name][utt.speaker_name].append(str(utt))
        for k, v in data.items():
            for s, utts in v.items():
                data[k][s] = sorted(utts)
        return data

    def cmvn_scp_data(self) -> Dict[str, Dict[str, str]]:
        """
        Generate the job's data for Kaldi's CMVN scp files

        Returns
        -------
        dict[str, dict[str, str]]
            Speaker to CMVN mapping, per dictionary name
        """
        data = {x: {} for x in self.current_dictionary_names}
        for s in self.current_speakers:
            if s.dictionary is None:
                key = None
            else:
                key = s.dictionary_name
            if s.cmvn:
                data[key][s.name] = s.cmvn
        return data

    def segments_scp_data(self) -> Dict[str, List[Any]]:
        """
        Generate the job's data for Kaldi's segments scp files

        Returns
        -------
        dict[str, list[Any]]
            Utterance to segment mapping, per dictionary name
        """
        data = {}
        for utt in self.current_utterances:
            if not utt.is_segment:
                continue
            data[utt.name] = utt.segment_for_scp()
        return data

    def construct_path_dictionary(
        self, directory: str, identifier: str, extension: str
    ) -> Dict[str, str]:
        """
        Helper function for constructing dictionary-dependent paths for the Job

        Parameters
        ----------
        directory: str
            Directory to use as the root
        identifier: str
            Identifier for the path name, like ali or acc
        extension: str
            Extension of the path, like .scp or .ark

        Returns
        -------
        dict[str, str]
            Path for each dictionary
        """
        output = {}
        for dict_name in self.current_dictionary_names:
            output[dict_name] = os.path.join(
                directory, f"{identifier}.{dict_name}.{self.name}.{extension}"
            )
        return output

    def construct_path(self, directory: str, identifier: str, extension: str) -> str:
        """
        Helper function for constructing dictionary-dependent paths for the Job

        Parameters
        ----------
        directory: str
            Directory to use as the root
        identifier: str
            Identifier for the path name, like ali or acc
        extension: str
            Extension of the path, like .scp or .ark

        Returns
        -------
        str
            Path
        """
        return os.path.join(directory, f"{identifier}.{self.name}.{extension}")

    def construct_dictionary_dependent_paths(
        self, directory: str, identifier: str, extension: str
    ) -> Dict[str, str]:
        """
        Helper function for constructing paths that depend only on the dictionaries of the job, and not the job name itself.
        These paths should be merged with all other jobs to get a full set of dictionary paths.

        Parameters
        ----------
        directory: str
            Directory to use as the root
        identifier: str
            Identifier for the path name, like ali or acc
        extension: str
            Extension of the path, like .scp or .ark

        Returns
        -------
        dict[str, str]
            Path for each dictionary
        """
        output = {}
        for dict_name in self.current_dictionary_names:
            output[dict_name] = os.path.join(directory, f"{identifier}.{dict_name}.{extension}")
        return output

    @property
    def dictionary_count(self):
        """Number of dictionaries currently used"""
        if self.subset_dictionaries:
            return len(self.subset_dictionaries)
        return len(self.dictionaries)

    @property
    def current_dictionaries(self) -> Collection[PronunciationDictionaryMixin]:
        """Current dictionaries depending on whether a subset is being used"""
        if self.subset_dictionaries:
            return [x for x in self.dictionaries if x.name in self.subset_dictionaries]
        return self.dictionaries

    @property
    def current_dictionary_names(self) -> List[Optional[str]]:
        """Current dictionary names depending on whether a subset is being used"""
        if self.subset_dictionaries:
            return sorted(self.subset_dictionaries)
        if self.dictionaries == {None}:
            return [None]
        return sorted(x.name for x in self.dictionaries)

    @property
    def current_speakers(self) -> Generator[Speaker]:
        """Current dictionary names depending on whether a subset is being used"""

        return self.speakers.subset(self.subset_speakers)

    @property
    def current_utterances(self) -> Generator[Utterance]:
        """Current dictionary names depending on whether a subset is being used"""
        for s in self.current_speakers:
            for u in s.utterances.subset(self.subset_utts):
                yield u

    def word_boundary_int_files(self) -> Dict[str, str]:
        """
        Generate mapping for dictionaries to word boundary int files

        Returns
        -------
        dict[str, str]
            Per dictionary word boundary int files
        """
        data = {}
        for dictionary in self.current_dictionaries:
            data[dictionary.name] = os.path.join(dictionary.phones_dir, "word_boundary.int")
        return data

    def output_for_features(self, split_directory: str) -> None:
        """
        Output the necessary files for Kaldi to generate features

        Parameters
        ----------
        split_directory: str
            Split directory for the corpus
        """
        wav_scp_path = self.construct_path(split_directory, "wav", "scp")
        if not os.path.exists(wav_scp_path):
            wav_scp = self.wav_scp_data()
            output_mapping(wav_scp, wav_scp_path, skip_safe=True)

        segments_scp_path = self.construct_path(split_directory, "segments", "scp")
        if not os.path.exists(segments_scp_path):
            segments = self.segments_scp_data()
            output_mapping(segments, segments_scp_path)

    def output_to_directory(self, split_directory: str) -> None:
        """
        Output job information to a directory

        Parameters
        ----------
        split_directory: str
            Directory to output to
        """

        spk2utt = self.spk2utt_scp_data()
        for dict_name, scp in spk2utt.items():
            spk2utt_scp_path = os.path.join(
                split_directory, f"spk2utt.{dict_name}.{self.name}.scp"
            )

            output_mapping(scp, spk2utt_scp_path)

        feats = self.feat_scp_data()
        for dict_name, scp in feats.items():
            feats_scp_path = os.path.join(split_directory, f"feats.{dict_name}.{self.name}.scp")
            output_mapping(scp, feats_scp_path)

        cmvn = self.cmvn_scp_data()
        for dict_name, scp in cmvn.items():
            cmvn_scp_path = os.path.join(split_directory, f"cmvn.{dict_name}.{self.name}.scp")
            output_mapping(scp, cmvn_scp_path)

        utt2spk = self.utt2spk_scp_data()
        for dict_name, scp in utt2spk.items():
            utt2spk_scp_path = os.path.join(
                split_directory, f"utt2spk.{dict_name}.{self.name}.scp"
            )
            output_mapping(scp, utt2spk_scp_path)

        text_scp = self.text_scp_data()
        for dict_name, scp in text_scp.items():
            if not scp:
                continue
            text_scp_path = os.path.join(split_directory, f"text.{dict_name}.{self.name}.scp")
            output_mapping(scp, text_scp_path, skip_safe=True)

        text_int = self.text_int_scp_data()
        for dict_name, scp in text_int.items():
            if dict_name is None:
                continue
            if not scp:
                continue
            text_int_scp_path = os.path.join(
                split_directory, f"text.{dict_name}.{self.name}.int.scp"
            )
            output_mapping(scp, text_int_scp_path, skip_safe=True)
